Object subclass:#SignalHandlingExample
         instanceVariableNames:''
         classVariableNames:'PrivateSignal Flag'
         poolDictionaries:''
         category:'Examples-Exceptions'!

"
 these examples show how Signals & Exceptions are used and handled.
 All kinds of signals (private, arithmetic and external; synchronous and
 asynchronous) are covered in these examples.

 try to understand these - they provide a powerful mechanism;
 (but some are a bit tricky :-)


 example1  - signal handler resumes i.e. continues execution

 example2  - same, but handler returns a value


 example3  - signal handler returns i.e. aborts execution

 example4  - same, but handler returns a value


 example5  - signal handler rejects i.e. passes control to next
             handler (in this example there is none, so Debugger
             is called)

 example6  - nested handlers, first rejects, 2nd returns

 example6b - same, but handler1 rejects by falling through


 example7  - handler does some cleanup, then restarts evaluation

 example8  - shows how to handle multiple signals with one handler

 example9  - shows how to handle all signals with one handler

 example10 - shows how to handle multiple arithmetic signals at once

 example11 - shows non nested signal handling by a handlerBlock

 example12 - same, but signal is raised by another process at later time

 example13 - a handler block and a handler context which takes it

 example14 - a handler block and a handler context which rejects 

 example15 - a handler block and a handler context which falls off

 example16 - fork another unix process, which sends us a signal
             (which is not handled)

 example17 - fork another unix process, which sends us a signal
             (which IS handled)

 example18 - signal a semaphore on arrival of a Unix signal
"!

!SignalHandlingExample class methodsFor:'examples'!

example1
    "the handler will proceed, thus continuing in the raising 
     method and returning its normal value"

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler; now resuming ..'.

            ex proceed
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example1"
!

example2
    "the handler will proceed, thus continuing in the raising method 
     and returning its normal value"

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler; now resuming ..'.

            ex proceedWith:'handlers value'
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example2"
!

example3
    "the handler will return, thus aborting the raising method 
     and returning nil from the computation"

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler; now returning ..'.

            ex return
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example3"
!

example4
    "the handler will return with a value, thus aborting the raising method 
     and returning the handlers value from the computation"

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler; now returning ..'.

            ex returnWith:'handlers value'
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example4"
!

example5
    "the handler will reject - thus leading into the debugger"

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler; rejecting'.

            ex reject
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example5"
!

example6
    "two handlers; the inner handler will reject - thus leading to the outer
     handler; the outer handler returns."

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler1; returning ..'.

            ex return
        ] do:[
            self exampleHandler
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example6"
!

example6b
    "two handlers; the inner handler falls through - thus leading to the outer
     handler; the outer handler returns."

    |val|

    PrivateSignal := Signal new.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler1; returning ..'.

            ex return
        ] do:[
            self exampleHandler2
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example6b"
!

example7
    "the handler will do some cleanup action (setting Flag) and restart the
     computation (which will not fail the second time)."

    |val|

    PrivateSignal := Signal new.
    Flag := false.
    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler1; cleanup and restart ..'.

            Flag := true.
            ex restart
        ] do:[
            self trapIfFlagIsFalse
        ].

    Transcript showCr:'final return value: ' , val printString

    "SignalHandlingExample example7"
!

example8
    "shows how to handle multiple signals (a SignalSet) with one handler.
     The handler just prints what happens (i.e. what type of signal) and
     proceeds."

    |val signals|

    PrivateSignal := Signal new notifierString:'my private signal'.
    signals := SignalSet with:PrivateSignal
                         with:Float divisionByZeroSignal.

    val :=
        signals handle:[:ex |
            Transcript show:'in handler, got signal: '.
            Transcript showCr:(ex signal notifierString).
            ex proceed
        ] do:[
            Transcript showCr:'starting ..'.
            self exampleHelper.
            Transcript showCr:'now dividing by zero ...'.
            self badDivision:0.
            Transcript showCr:'now at end of block'
        ].

    Transcript showCr:'thats it'

    "SignalHandlingExample example8"
!

example9
    "shows how to handle multiple signals (a SignalSet containing anySignal)
     with one handler.
     The handler just prints what happens (i.e. what type of signal) and
     proceeds."

    |val|

    PrivateSignal := Signal new notifierString:'my private signal'.

    SignalSet anySignal handle:[:ex |
        Transcript show:'in handler, got signal: '.
        Transcript showCr:ex signal notifierString.
        ex proceed
    ] do:[
        Transcript showCr:'now raising my signal ..'.
        self exampleHelper.
        Transcript showCr:'now dividing by zero ...'.
        self badDivision:0.
        Transcript showCr:'now sending something not understood ...'.
        self perform:#fooBarBaz.
        Transcript showCr:'now at end of block'
    ]

    "SignalHandlingExample example9"
!

example10
    "shows how to handle arithmetic signals with one handler.
     The handler just prints what happens (i.e. what type of signal) and
     proceeds."

    |val|

    (Float anyArithmeticSignal) handle:[:ex |
        Transcript show:'in handler, got signal: '.
        Transcript show:ex signal notifierString.
        Transcript showCr:'; continue'.
        ex proceed
    ] do:[
        Transcript showCr:'1 ..'.
        self computation1:0.   "raises a domain error"
        Transcript showCr:'2 ...'.
        self computation2:5.   "raises a domain error"
        Transcript showCr:'3 ...'.
        self computation3:0.   "raises a division by zero"
        Transcript showCr:'now at end of block'
    ]

    "SignalHandlingExample example10"
!

example11
    "shows non nested signal handling.
     The handling block shows a warnBox and proceeds."

    |val mySignal|

    mySignal := Signal new.
    mySignal handlerBlock:[:ex :con |
        self warn:'a private Signal occured'.
    ].

    mySignal raise.

    "SignalHandlingExample example11"
!

example12
    "shows non nested signal handling.
     The handling block shows a warnBox and proceeds."

    |val mySignal|

    mySignal := Signal new.
    mySignal handlerBlock:[:ex :con |
        self warn:'a private Signal occured'.
    ].

    [
        (Delay forSeconds:3) wait.
        mySignal raise.
    ] fork.

    "SignalHandlingExample example12"
!

example13
    "shows non nested signal / nested signal handling.
     The handling block shows a warnBox and proceeds, while
     a handlercontext handles the signal by returning"

    |val mySignal|

    mySignal := Signal new.
    mySignal handlerBlock:[:ex :con |
        self warn:'a private Signal occured'.
    ].

    [
        (Delay forSeconds:3) wait.
        mySignal handle:[:ex |
            Transcript showCr:'handler took it'.
            ex return
        ] do:[
            (Delay forSeconds:1) wait.
            mySignal raise.
        ]
    ] fork.

    "SignalHandlingExample example13"
!

example14
    "shows non nested signal / nested signal handling.
     The handling block shows a warnBox and proceeds, while
     a handlercontext rejects the signal"

    |val mySignal|

    mySignal := Signal new.
    mySignal handlerBlock:[:ex :con |
        self warn:'a privateSignal occured'.
    ].

    [
        (Delay forSeconds:3) wait.
        mySignal handle:[:ex |
            Transcript showCr:'handler rejects ...'.
            ex reject
        ] do:[
            (Delay forSeconds:1) wait.
            mySignal raise.
        ]
    ] fork.

    "SignalHandlingExample example14"
!

example15
    "shows non nested signal / nested signal handling.
     The handling block shows a warnBox and proceeds, while
     a handlercontext handles the signal, but falls off"

    |val mySignal|

    mySignal := Signal new.
    mySignal handlerBlock:[:ex :con |
        self warn:'a privateSignal occured'.
    ].

    [
        (Delay forSeconds:3) wait.
        mySignal handle:[:ex |
            Transcript showCr:'handler falls through ...'.
        ] do:[
            (Delay forSeconds:1) wait.
            mySignal raise.
        ]
    ] fork.

    "SignalHandlingExample example15"
!

example16
    "getting a UNIX signal:
     fork a process, to send a signal after 3 seconds.
     The signal is not handled - thus leading into debugger."

    |mySignal command childId|

    "install a signal, which is not handled"
    mySignal := Signal new.
    Object operatingSystemSignal:(OperatingSystem sigUSR1) install:mySignal.

    "let them come ..."
    OperatingSystem enableSignal:(OperatingSystem sigUSR1).

    "fork off the process"
    command := 'sleep 3 ; kill -USR1 ' , OperatingSystem getProcessId printString.
    (childId := OperatingSystem fork) == 0 ifTrue:[
        "I am the child"
        OperatingSystem executeCommand:command.
        OperatingSystem exit.
    ].

    "father leaves ..."

    "SignalHandlingExample example16"
!

example17
    "handling a UNIX signal:
     fork a process, to send a signal after 3 seconds.
     the signal is handled."

    |mySignal command childId|

    "install a signal,"
    mySignal := Signal new.
    Object operatingSystemSignal:(OperatingSystem sigUSR1) install:mySignal.

    "which is handled"
    mySignal handlerBlock:[:ex :con |
        self warn:'a signal arrived'.
        "uninstall"
        Object operatingSystemSignal:(OperatingSystem sigUSR1) install:nil.
    ].

    "let them come ..."
    OperatingSystem enableSignal:(OperatingSystem sigUSR1).

    "fork off the process"
    command := 'sleep 3 ; kill -USR1 ' , OperatingSystem getProcessId printString.
    (childId := OperatingSystem fork) == 0 ifTrue:[
        "I am the child"
        OperatingSystem executeCommand:command.
        OperatingSystem exit.
    ].

    "father leaves ..."

    "SignalHandlingExample example17"
!

example18
    "trigger a semaphore on arrival of a UNIX signal:
     fork a process, to send a signal after 3 seconds.
     the signal is globally handled, by triggering a semaphore, on which we wait here"

    |USR1Semaphore mySignal command childId|

    "create the semaphore"
    USR1Semaphore := Semaphore new.

    "install a signal,"
    mySignal := Signal new.
    Object operatingSystemSignal:(OperatingSystem sigUSR1) install:mySignal.

    "which is handled - and wakes up my semaphore"
    mySignal handlerBlock:[:ex :con |
        Transcript showCr:'here signal handler; now trigger sema'.
        USR1Semaphore signal.
        "uninstall"
        Object operatingSystemSignal:(OperatingSystem sigUSR1) install:nil.
    ].

    "let them come ..."
    OperatingSystem enableSignal:(OperatingSystem sigUSR1).

    "fork off the process"
    command := 'sleep 3 ; kill -USR1 ' , OperatingSystem getProcessId printString.
    (childId := OperatingSystem fork) == 0 ifTrue:[
        "I am the child"
        OperatingSystem executeCommand:command.
        OperatingSystem exit.
    ].

    "I fork off a process to wait"

    [
        Transcript showCr:'process waits on sema ...'.
        USR1Semaphore wait.
        Transcript showCr:'I got the signal - Hello there !!'.
    ] fork.

    Transcript showCr:'father does whatever ...'.

    "SignalHandlingExample example18"
!

example19

    Object recursionInterruptSignal handle:[:ex |
        'oops' printNewline.
        ex return
    ] do:[
        'start ...' printNewline.
        10000 factorial.
        'done ...' printNewline.
    ].
    'thats it' printNewline

    "SignalHandlingExample example19"
!

"---------------------------------- no more examples, but helpers ---------------------------"

exampleHandler
    |val|

    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler2; rejecting ..'.

            ex reject
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'exampleHandler return value: ' , val printString
!

exampleHandler2
    |val|

    val :=
        PrivateSignal handle:[:ex |
            Transcript showCr:'in handler2; fall through ..'.
            'thats it'
        ] do:[
            self exampleHelper
        ].

    Transcript showCr:'exampleHandler return value: ' , val printString
!

exampleHelper
    |val|

    Transcript showCr:'in raising method - raising signal ..'.

    val := PrivateSignal raise.

    Transcript show:'in raising method after raise; '.
    Transcript showCr:'raise returned: ' , val printString.
    Transcript showCr:'return from raising method ...'.

    ^ 'normal value'
!

badDivision:arg
    ^ 3 / arg
!

trapIfFlagIsFalse
    Flag ifTrue:[
        Transcript showCr:'in processing method - all ok; return ..'.
        ^ 'normal value'
    ].

    Transcript showCr:'in processing method - raising signal ..'.

    PrivateSignal raise.

    "not reached if handler restarts ..."
    Transcript showCr:'return from raising method ...'.

    ^ 'normal value'
!

computation1:arg
    arg log      "will raise a domain error if arg is zero"
!

computation2:arg
    arg arcSin   "will raise a domain error for some arguments"
!

computation3:arg
    3.0 / arg    "will raise a divisionByZero signal for zero arg"
! !
