"
 COPYRIGHT (c) 1991 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#Image
         instanceVariableNames:'bytes width height 
                                bitsPerSample samplesPerPixel
                                colorMap photometric 
                                device deviceForm monoDeviceForm
                                fullColorDeviceForm'
         classVariableNames:'Lobby 
                             DitherAlgorithm NumberOfDitherColors
                             FileFormats'
         poolDictionaries:''
         category:'Graphics-Display Objects'
!

Image comment:'

COPYRIGHT (c) 1991 by Claus Gittinger
              All Rights Reserved

$Header: /claus/smalltalk/st2.6/libview/RCS/Image.st,v 1.11 94/02/25 13:12:22 claus Exp $
written Summer 91 by claus
'!

!Image class methodsFor:'documentation'!

documentation
"
    this class provides (some time in the future) representation
    for all kinds of images (monochrome, greyscale and color)
    and will finally replace Form - its still under construction.

    An Image keeps all info in a device independent way, but may get
    associated to a device. The data held keeps all information which
    was originally present, even if the display-device has lower resolution.
    Therefore, it is possible to process and manipulate images without loosing 
    color information - even on low color resolution displays.

    Usually, you get a device specific representation of the image by
    sending an image the 'on:aDevice' message, which will create
    a (possibly dithered) form, representing the image using the currently
    available colors.

    Sometimes, an explicit monochrome representation is needed 
    (X servers take monochrome icons only), this can be created by sending
    'anImage monochromeOn:aDevice'.

    Also, it is planned to generate another hi-color resolution version,
    which needs its own colormap to be installed and allows use of all
    256 colors on an 8bit display (not currently implemented).

    To convert pictures from/to external file-formats, readers are used
    which have the file-format knowledge (see TIFFReader, GIFReader etc.).

    The algorithms in here (especially dithering & color allocation) are
    experimental and far from being perfect (some are very slow). 
    Much more work is needed and will be done in the near future ...
    Dithering is done as:

       DitherAlgorithm:

       nil                  a simple threshold algorithm
                            (i.e. for mono, p<0.5 -> black, p>=0.5 -> white)

       #pattern             patterned dither
                            (for p, take dithered color to fill pixel;
                             uses dithering in color-class)

       #error               error diffusion dither (Floyd-Steinberg)
                            planned

    File formats are handled by subclasses of ImageReader, which understand
    a specific format. You can add more readers, by adding an association
    such as ('.jpg' -> JPEGReader) to the class variable 'FileFormats' (see
    Image initialize.

    instance variables:

        width               <Integer>       the width in pixels
        height              <Integer>       the height in pixels
        bytes               <ByteArray>     the full image information
        photometric         <Symbol>        #rgb, #palette, #blackIs0 or #whiteIs0
        samplesPerPixel     <Integer>       the number of planes
        bitsPerSample       <Array>         the number of bits per plane
        colorMap            <Array>         an Array of 3 arrays containing red,
                                            green and blue values (0..255 biased)
        device              <Workstation>   the device on which deviceForm,
                                            monoDeviceForm and lowResDeviceForm are
        deviceForm          <Form>          the device form which gives the best
                                            possible aproximation of the image on
                                            device using standard colors.
        monoDeviceForm      <Form>          the device form which gives a monochrome
                                            aproximation of the image on device.
        fullColorDeviceForm <Form>          the device form which gives the best
                                            possible aproximation of the image on
                                            device using private colors.

    class variables:

        Lobby               <Registry>      keeps track of known images
        DitherAlgorithm     <Symbol>        defines how to dither
        NumberOfDitherColors <Integer>      defines, how many dither colors to use
        FileFormats         <Dictionary>    associates filename extensions to
                                            image reader classes (now set-up in startup-file)
    todo:
        there is currently no mask stored/available in the image itself; currently masks
        have to be stored as separate bitmaps. (which is bad for image-file formats, which 
        provide a mask)
"
! !

!Image class methodsFor:'misc'!

dither:aSymbol
    "define how to dither - #pattern, #error or none;
     error diffusion dithering is currently not implemented,
     pattern dither is currently very slow."

    DitherAlgorithm := aSymbol

    "Image dither:#pattern"
    "Image dither:#error"
    "Image dither:nil"
!

numberOfDitherColors:n
    "define how many colors (i.e. patterns) to use when
     doing a pattern dither (good values are:)"

    NumberOfDitherColors := n
! !

!Image class methodsFor:'initialization'!

initialize
    "initialize class constants"

    "setup tracker of known pictures"
    Lobby isNil ifTrue:[
        Lobby := Registry new.
        ObjectMemory addDependent:self.
    ].

    "define algorithm to use for dithering - currently only nil or #pattern supported"
    DitherAlgorithm := #pattern.   "will be changed to error as soon as implemented"
    (Display notNil and:[Display hasGreyscales]) ifFalse:[
        NumberOfDitherColors := 64
    ] ifTrue:[
        "as far as I remember, this is about the number of grey values, the eye can distinguish"
        NumberOfDitherColors := 100
    ].

    "define reader classes"
    FileFormats isNil ifTrue:[
        self initializeFileFormatTable
    ].
!

initializeFileFormatTable
    "initialize a default table to map from file extension to reader class.
     The mapping here is a default needed for proper operation of ST/X;
     see the 'smalltalk.rc' startup file for a real (full) map."

    FileFormats := Dictionary new.
    FileFormats at:'.xbm'  put:XBMReader.
    FileFormats at:'.tiff' put:TIFFReader.
    FileFormats at:'.gif'  put:GIFReader.
    FileFormats at:'.img'  put:IMGReader.
    FileFormats at:'.icon' put:SunRasterReader.

    "Image initializeFileFormatTable"
!

fileFormats
    "return the collection of supported file formats.
     The returned dictionary maps file-extensions to image reader classes."

    ^ FileFormats

    "Image fileFormats"
!

flushDeviceImages
    "simply unassign all pictures from their device"

    Lobby contentsDo:[:anImage |
        anImage restored
    ]
!

update:something
    "flush all device specific stuff when restarted from a snapshot"

    (something == #restarted) ifTrue:[
        self flushDeviceImages
    ]
! !

!Image class methodsFor:'screen capture'!

fromScreen
    "return an image of the full screen"

    ^ self fromScreen:(0@0 corner:(Display width@Display height))
!

fromScreenArea
    "return an image of a part of the screen; 
     let user specify screen area."

    ^ self fromScreen:(Rectangle fromUser)

    "Image fromScreenArea"
!

fromScreen:aRectangle
    "return an image of a part of the screen"

    ^ self fromScreen:aRectangle on:Display

    "Image fromScreen:(0@0 corner:100@100)"
!

fromScreen:aRectangle on:aDisplay
    "return an image of a part of the screen, which may be on
     another Display."

    |depth img|

    depth := aDisplay depth.
    img := (self implementorForDepth: depth) new.
    ^ img fromScreen:aRectangle on:aDisplay

    "Image fromScreen:(0@0 corner:100@100)"
!

fromView:aView
    "return an image taken from a view"

    |org dev|

    dev := aView device.
    org := dev translatePoint:(0@0)
                         from:(aView id)
                           to:(DisplayRootView on:dev) id.
    ^ self fromScreen:(org extent:aView extent) on:dev

    "Image fromView:(Launcher allInstances first topView)"
    "Image fromView:(SystemBrowser allInstances first topView)"
! !

!Image class methodsFor:'reading from file'!

fromFile:aFileName
    "read an image from a file - this methods tries to find
     out the file format itself (by the extension and by contents)
     and lets the appropriate reader read the file."

    |readerClass image name|

    name := aFileName.
    (name endsWith:'.Z') ifTrue:[
        name := name copyTo:(name size - 2)
    ].
    FileFormats associationsDo:[:assoc |
        (name endsWith:(assoc key)) ifTrue:[
            readerClass := assoc value.
            readerClass notNil ifTrue:[
                image := readerClass fromFile:aFileName.
                image notNil ifTrue:[^ image].
            ]
        ]
    ].
    "no known extension - ask all readers if they know
     this format ..."

    FileFormats associationsDo:[:assoc |
        readerClass := assoc value.
        readerClass notNil ifTrue:[
            (readerClass isValidImageFile:aFileName) ifTrue:[
                ^ readerClass fromFile:aFileName
            ]
        ]
    ].

    "nope - unknown format"
    'unknown image file format: ' errorPrint. aFileName errorPrintNewline.
    ^ nil

    "Image fromFile:'bitmaps/dano.tiff'"
    "Image fromFile:'bitmaps/test.fax'"
    "Image fromFile:'bitmaps/voice.tiff'"
    "Image fromFile:'voice.tiff'"

    "Image fromFile:'../fileIn/bitmaps/claus.gif'"
    "Image fromFile:'../fileIn/bitmaps/garfield.gif'"

    "Image fromFile:'../fileIn/bitmaps/founders.im8'"
    "Image fromFile:'../goodies/faces/next.com/steve.face'"

    "Image fromFile:'/LocalLibrary/Images/OS2/dos3.ico'"
    "Image fromFile:'bitmaps/globe1.xbm'"
    "Image fromFile:'bitmaps/globe1.xbm.Z'"
    "Image fromFile:'bitmaps/hello_world.icon'"
! !

!Image class methodsFor:'queries'!

implementorForDepth: depth
    "return the class, which best implements images of depth"

    depth == 1 ifTrue:[^ Depth1Image].
    depth == 2 ifTrue:[^ Depth2Image].
    depth == 4 ifTrue:[^ Depth4Image].
    depth == 8 ifTrue:[^ Depth8Image].
    depth == 24 ifTrue:[^ Depth24Image].
    ^ self
! !

!Image methodsFor:'instance release'!

restored
    "flush device specifics after a snapin or binary restore"

    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.
    Lobby unregister:self
!

disposed
    "some Image has been collected - nothing to do"
! !

!Image methodsFor:'inpecting'!

inspect
    "redefined to launch an ImageInspector on the receiver
     (instead of the default InspectorView)."

    ImageInspectorView isNil ifTrue:[
        super inspect
    ] ifFalse:[
        ImageInspectorView openOn:self
    ]
! !

!Image methodsFor:'accessing'!

device
    "return the device, the receiver is associated with.
     Return nil, if the image is unassigned."

    ^ device
!

id
    "return the id of the image on the device.
     Return nil, if the image is unassigned."

    deviceForm isNil ifTrue:[^ nil].
    ^ deviceForm id
!

monochromeId
    "return the id of the monochrome image on the device.
     Return nil, if the image is unassigned."

    monoDeviceForm isNil ifTrue:[^ nil].
    ^ monoDeviceForm id
!

fullColorId
    "return the id of the full color image on the device.
     Return nil, if the image is unassigned."

    fullColorDeviceForm isNil ifTrue:[^ nil].
    ^ fullColorDeviceForm id
!

width
    "return the width of the image"

    ^ width
!

width:aNumber
    "set the width of the image"

    width := aNumber
!

height
    "return the height of the image"

    ^ height
!

height:aNumber
    "set the height of the image"

    height := aNumber
!

extent
    "return the images extent"

    ^ width@height
!

samplesPerPixel
    "return the number of samples per pixel in the image.
     The return value is an array of bits-per-plane."

    ^ samplesPerPixel
!

samplesPerPixel:aNumber
    "set the array of samples per pixel"

    samplesPerPixel := aNumber
!

depth
    "return the depth of the image"

    ^ self bitsPerPixel
!

bitsPerSample
    "return the number of bits per sample.
     The return value is an array of bits-per-plane."

    ^ bitsPerSample
!

bitsPerSample:aNumber
    bitsPerSample := aNumber
!

bitsPerPixel
    "return the number of bits per pixel"

    ^ (bitsPerSample inject:0 into:[:sum :i | sum + i])
!

bitsPerRow
    "return the number of bits in one scanline of the image"

    ^  width * (self bitsPerPixel).
!

bytesPerRow
    "return the number of bytes in one scanline of the image"

    |bitsPerRow bytesPerRow|

    bitsPerRow := width * (self bitsPerPixel).
    bytesPerRow := bitsPerRow // 8.
    ((bitsPerRow \\ 8) ~~ 0) ifTrue:[
        bytesPerRow := bytesPerRow + 1
    ].
    ^ bytesPerRow
!

data
    "for backward compatibility - will vanish"

    ^ bytes
!

data:aByteArray
    "for backward compatibility - will vanish"

    bytes := aByteArray
!

bits:aByteArray
    "set the raw data"

    bytes := aByteArray
!

bits
    "return the raw image data; depending on the photometric,
     this has to be interpreted as monochrome, greyscale,
     palette or rgb data. It is also packed to be dense, so
     a 4 bitPerSample palette image will store 2 pixels per byte."

    ^ bytes
!

photometric
    "return the photometric, a symbol such as #palette, #rgb etc."

    ^ photometric
!

photometric:aSymbol
    "set the photometric"

    photometric := aSymbol
!

colorMap:anArrayOfArrays
    colorMap := anArrayOfArrays
!

colorMap
    "return the colormap"

    ^ colorMap
!

at:aPoint
    "retrieve the pixel at aPoint; return a color.
     Pixels start at 0@0 for upper left pixel, end at
     width-1@height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self atX:aPoint x y:aPoint y
!

valueAt:aPoint
    "retrieve the pixel at aPoint; return a pixel value.
     Pixels start at 0@0 for upper left pixel, end at
     width-1@height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self valueAtX:aPoint x y:aPoint y
!

atX:x y:y
    "retrieve a pixel at x/y; return a color.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self subclassResponsibility
!

valueAtX:x y:y
    "retrieve the pixel at aPoint; return a pixel value.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     width-1@height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self subclassResponsibility
!

at:aPoint put:aColor
    "set the pixel at aPoint to aColor.
     Pixels start at 0@0 for upper left pixel, end at
     width-1@height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self atX:aPoint x y:aPoint y put:aColor
!

atX:x y:y put:aColor
    "set the pixel at x/y to aColor.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self subclassResponsibility
!

atX:x y:y putValue:aPixelValue
    "set the pixel at x/y to aPixelValue.
     Pixels start at x=0 , y=0 for upper left pixel, end at
     x = width-1, y=height-1 for lower right pixel.
     You should not use this method for image-processing, its
     very slow ..."

    ^ self subclassResponsibility
!

atY:y from:x1 to:x2 do:aBlock
    "perform aBlock for each pixel from x1 to x2 in row y.
     The block is passed the color at each pixel.
     This method allows slighly faster processing of an
     image than using atX:y:, since some processing can be 
     avoided when going from pixel to pixel. However, for 
     real image processing, specialized methods should be written."

    ^ self subclassResponsibility.
! !

!Image methodsFor:'storing'!

storeOn:aStream
    aStream nextPutAll:'(' , self class name , ' new)'.
    aStream nextPutAll:' width: '. width storeOn:aStream.
    aStream nextPutAll:'; height: '. height storeOn:aStream.
    aStream nextPutAll:'; photometric:('. photometric storeOn:aStream.
    aStream nextPutAll:'); bitsPerSample:('. bitsPerSample storeOn:aStream.
    aStream nextPutAll:'); samplesPerPixel:('. samplesPerPixel storeOn:aStream. 
    aStream nextPutAll:'); bits:('. bytes storeOn:aStream. aStream nextPutAll:'); '.
    colorMap notNil ifTrue:[
        aStream nextPutAll:'; colorMap:('.
        colorMap storeOn:aStream.
        aStream nextPutAll:')'.
    ].
    aStream nextPutAll:' yourself'
! !

!Image methodsFor:'screen capture'!

fromScreen:aRectangle
    "read an image from the display screen"

    ^ self fromScreen:aRectangle on:Display
!

fromScreen:aRectangle on:aDevice
    "read an image from aDevices display screen.
     Since I have no other displays, only the MonoChrome, StaticGrey 
     and PseudoColor cases have been tested ... 
     (especially True- and DirectColor may be wrong)"

    |depth visType root 
     x        "{ Class: SmallInteger }"
     y        "{ Class: SmallInteger }"
     w        "{ Class: SmallInteger }"
     h        "{ Class: SmallInteger }"
     dstIndex "{ Class: SmallInteger }" 
     srcIndex "{ Class: SmallInteger }" 
     inData tmpData usedColors nUsed 
     rMap gMap bMap bitsPerPixel bytesPerLine
     info bytesPerLineIn curs cid|

    curs := Cursor sourceForm:(Form fromFile:'Camera.xbm')
                     maskForm:(Form fromFile:'Camera_m.xbm')
                      hotSpot:16@16.
    curs notNil ifTrue:[
        cid := (curs on:aDevice) id
    ].
    "actually have to grabServer ... but thats not yet available"
    ActiveGrab := DisplayRootView on:aDevice.
    aDevice grabPointerIn:(DisplayRootView on:aDevice) id
               withCursor:cid pointerMode:#async keyboardMode:#sync confineTo:nil.

    visType := aDevice visualType.
    depth := aDevice depth.
    (visType == #StaticGray) ifTrue:[
        (aDevice blackpixel == 0) ifTrue:[
            photometric := #blackIs0
        ] ifFalse:[
            photometric := #whiteIs0
        ].
        samplesPerPixel := 1.
        bitsPerPixel := depth.
        bitsPerSample := Array with:bitsPerPixel.
    ] ifFalse:[
        ((visType == #PseudoColor) or:[(visType == #StaticColor) or:[visType == #GrayScale]]) ifTrue:[
            photometric := #palette.
            samplesPerPixel := 1.
            bitsPerPixel := depth.
            bitsPerSample := Array with:bitsPerPixel.
        ] ifFalse:[
            ((visType == #TrueColor) or:[visType == #DirectColor]) ifTrue:[
                photometric := #rgb.
                samplesPerPixel := 3.
                bitsPerPixel := 24.
                bitsPerSample := #(8 8 8)
            ] ifFalse:[
                self error:'screen visual not supported'.
                ^ nil
            ]
        ]
    ].

    "dont know yet, how display pads; assume worst case, 
     offering enough space for 32 bit padding"

    w := width := aRectangle width.
    h := height := aRectangle height.
    x := aRectangle left.
    y := aRectangle top.

    bytesPerLine := (w * bitsPerPixel + 31) // 32 * 4.
    inData := ByteArray uninitializedNew:(bytesPerLine * height).

    root := DisplayRootView on:aDevice.
    info := aDevice getBitsFrom:root id x:x y:y width:w height:h into:inData. 

    "if the padding is not correct .."
    bytesPerLineIn := (info at:3).                                "what I got"
    bytesPerLine := (w * bitsPerPixel + 7) // 8.                  "what I want"
    (bytesPerLine ~~ bytesPerLineIn) ifTrue:[
        tmpData := inData.
        inData := ByteArray uninitializedNew:(bytesPerLine * height).
        srcIndex := 1.
        dstIndex := 1.
        1 to:height do:[:h |
            inData replaceFrom:dstIndex to:(dstIndex + bytesPerLine - 1)
                          with:tmpData startingAt:srcIndex.
            dstIndex := dstIndex + bytesPerLine.
            srcIndex := srcIndex + bytesPerLineIn
        ]
    ] ifFalse:[
        (bytesPerLine * height) ~~ inData size ifTrue:[
            tmpData := inData.
            inData := ByteArray uninitializedNew:(bytesPerLine * height).
            inData replaceFrom:1 to:bytesPerLine * height with:tmpData startingAt:1
        ]
    ].

    "info printNewline. "

    ((visType == #StaticGray) or:[visType == #TrueColor]) ifTrue:[
        "were done, the pixel values are the rgb/grey values"
        bytes := inData.
        ^ self
    ].

    "what we have now are the color numbers - still need the r/g/b values"
    "find out, which colors are in the picture"

    usedColors := inData usedValues.
    nUsed := usedColors maximumValue + 1.

    "get the palette"
    rMap := Array new:nUsed.
    gMap := Array new:nUsed.
    bMap := Array new:nUsed.
    usedColors do:[:colorIndex |
        |i scale|

        i := colorIndex + 1.
        scale := 255.0 / 100.0.
        aDevice getRGBFrom:colorIndex into:[:r :g :b |
            rMap at:i put:(r * scale) rounded.
            gMap at:i put:(g * scale) rounded.
            bMap at:i put:(b * scale) rounded
        ]
    ].
    colorMap := Array with:rMap with:gMap with:bMap.
    bytes := inData.

    aDevice ungrabPointer.
    ActiveGrab := nil.

    "Image new fromScreen:((0 @ 0) corner:(100 @ 100)) on:Display"
    "Image new fromScreen:((0 @ 0) corner:(500 @ 500)) on:Display"
! !

!Image methodsFor:'saving on file'!

saveOn:aFileName
    "save the image in a aFileName. The suffix of the filename
     controls the format. Currently, not all formats may be supported
     (see ImageReader subclasses implementing save:onFile:)"

    FileFormats associationsDo:[:a |
        (aFileName endsWith:(a key)) ifTrue:[
            ^ (a value) save:self onFile:aFileName
        ]
    ].
    "no known extension - could ask user for the format here.
     currently default to tiff format."

    'unknown extension - could not figure out format - using tiff' errorPrintNewline.
    ^ self saveOn:aFileName using:TIFFReader
!

saveOn:aFileName using:readerClass
    "save the receiver using the representation class"

    readerClass save:self onFile:aFileName

    "anImage saveOn:'myImage' using:TIFFReader"
    "anImage saveOn:'myImage' using:XBMReader"
! !

!Image methodsFor:'converting'!

on:aDevice
    "make the image device dependent for aDevice"

    ((aDevice == device) and:[deviceForm notNil]) ifTrue:[^ self].
    deviceForm := self asFormOn:aDevice.
    deviceForm notNil ifTrue:[
        device isNil ifTrue:[
            device := aDevice.
            Lobby register:self
        ] ifFalse:[
            device := aDevice.
            Lobby changed:self
        ].
        deviceForm forgetBits
    ]
!

monochromeOn:aDevice
    "return a monochrome device image of the receiver for aDevice.
     (monochrome, even if device supports colors)"

    ((aDevice == device) and:[monoDeviceForm notNil]) ifTrue:[^ self].
    monoDeviceForm := self asMonochromeFormOn:aDevice.
    monoDeviceForm notNil ifTrue:[
        device := aDevice.
        monoDeviceForm forgetBits
    ]
!

asFormOn:aDevice
    "get a device form, with best possible approximation"

    (photometric == #palette) ifTrue:[
        ^ self paletteImageAsFormOn:aDevice
    ].

    (photometric == #rgb) ifTrue:[
        ^ self rgbImageAsFormOn:aDevice
    ].

    ^ self greyImageAsFormOn:aDevice
!

asMonochromeFormOn:aDevice
    "get a monochrome device form"

    (photometric == #palette) ifTrue:[
        ^ self paletteImageAsMonoFormOn:aDevice
    ].

    (photometric == #rgb) ifTrue:[
        ^ self rgbImageAsMonoFormOn:aDevice
    ].

    ^ self greyImageAsMonoFormOn:aDevice
! !

!Image methodsFor:'converting rgb images'!

rgbImageAsFormOn:aDevice
    "convert am rgb image to a device form form aDevice.
     Return the device form."

    |visual|

    visual := aDevice visualType.
    (visual == #StaticGray) ifTrue:[
        ^ self rgbImageAsGreyFormOn:aDevice
    ].
    (visual == #TrueColor) ifTrue:[
        ^ self rgbImageAsTrueFormOn:aDevice
    ].
    ^ self rgbImageAsPseudoFormOn:aDevice
!

rgbImageAsGreyFormOn:aDevice
    "convert an rgb image to a grey image for greyscale displays"

    |deviceDepth|

    deviceDepth := aDevice depth.

    "I have specially tuned methods for monochrome"
    (deviceDepth == 1) ifTrue:[
        DitherAlgorithm == #error ifTrue:[
            ^ self rgbImageAsErrorDitheredGreyFormOn:aDevice
        ].
        DitherAlgorithm == #pattern ifTrue:[
            ^ self rgbImageAsPatternDitheredGreyFormOn:aDevice
        ].
        ^ self rgbImageAsMonoFormOn:aDevice
    ].

    "and for 2plane greyscale (i.e. NeXTs)"
    (deviceDepth == 2) ifTrue:[
        DitherAlgorithm == #error  ifTrue:[
            ^ self rgbImageAsErrorDitheredGreyFormOn:aDevice
        ].
        DitherAlgorithm == #pattern  ifTrue:[
            ^ self rgbImageAsPatternDitheredGreyFormOn:aDevice
        ].
        ^ self rgbImageAs2PlaneFormOn:aDevice
    ].

    (deviceDepth == 8) ifTrue:[
        ^ self rgbImageAs8BitGreyFormOn:aDevice
    ].

    "mhmh need another converter ...
     till then we do:"
    DitherAlgorithm == #error  ifTrue:[
        ^ self rgbImageAsErrorDitheredGreyFormOn:aDevice
    ].
    DitherAlgorithm == #pattern  ifTrue:[
        ^ self rgbImageAsPatternDitheredGreyFormOn:aDevice
    ].
    ^ self rgbImageAsMonoFormOn:aDevice
!

rgbImageAsMonoFormOn:aDevice
    "return a 1-bit monochrome form for aDevice from the rgb picture,
     using a threshold algorithm. 
     (i.e. grey value < 0.5 -> black, grey value >= 0.5 -> white)."

    ^ self subclassResponsibility
!

rgbImageAsPatternDitheredGreyFormOn:aDevice
    "return a dithered greyForm for aDevice from the palette picture.
     works for any destination depth.
     A slow algorithm, using draw into the form (which indirectly does
     the dither) - should be rewritten."

    ^ self subclassResponsibility
!

rgbImageAs2PlaneFormOn:aDevice
    "return a 2-bit device form for aDevice from the rgb picture,
     using a threshold algorithm. 
     (i.e. grey value < 0.25 -> black // 0.25..0.5 -> darkgrey //
      0.5 .. 0.75 -> lightgrey // > 0.75 -> white)."

    ^ self subclassResponsibility
!

rgbImageAs8BitGreyFormOn:aDevice
    "return an 8-bit greyForm from the rgb picture"

    ^ self subclassResponsibility
!

rgbImageAsPseudoFormOn:aDevice
    "return a pseudocolor form from the rgb-picture"

    ^ self subclassResponsibility
! !

!Image methodsFor:'converting palette images'!

paletteImageAsFormOn:aDevice
    (aDevice visualType == #StaticGray) ifTrue:[
        (aDevice depth == 8) ifTrue:[
            ^ self paletteImageAsGreyFormOn:aDevice
        ].

        DitherAlgorithm == #pattern ifTrue:[
            ^ self paletteImageAsPatternDitheredGreyFormOn:aDevice
        ].

        (aDevice depth == 2) ifTrue:[
            ^ self paletteImageAs2PlaneFormOn:aDevice
        ].

        ^ self paletteImageAsMonoFormOn:aDevice
    ].
    ^ self paletteImageAsPseudoFormOn:aDevice
!

paletteImageAsMonoFormOn:aDevice
    "return a 1-bit monoForm from the palette picture"

    ^ self subclassResponsibility
!

paletteImageAs2PlaneFormOn:aDevice
    "return a 2-bit greyForm from the palette picture"

    ^ self subclassResponsibility
!

paletteImageAsPseudoFormOn:aDevice
    "return a pseudoForm from the palette picture."

    ^ self subclassResponsibility
!

paletteImageAsGreyFormOn:aDevice
    "return an 8-bit greyForm from the palette picture"

    ^ self subclassResponsibility
!

paletteImageAsPatternDitheredGreyFormOn:aDevice
    "return a dithered greyForm from the palette picture."

    ^ self subclassResponsibility
! !

!Image methodsFor:'converting greyscale images'!

greyImageAsFormOn:aDevice
    "return a thresholded greyForm from the grey picture."

    |pictureDepth nPlanes f|

    nPlanes := samplesPerPixel.
    (nPlanes == 2) ifTrue:[
        'alpha plane ignored' errorPrintNewline.
        nPlanes := 1
    ].

    pictureDepth := bitsPerSample at:1.

    "monochrome is very easy ..."

    (pictureDepth == 1) ifTrue:[
        ^ Form width:width height:height fromArray:bytes on:aDevice
    ].

    (aDevice visualType == #StaticGray) ifTrue:[
        (aDevice depth == pictureDepth) ifTrue:[

            "greyscale is easy, if the depths match"

            f := Form width:width height:height depth:pictureDepth on:aDevice.
            f isNil ifTrue:[^ nil].
            f initGC.

            "if device has white at the opposite corner ..."
            ((aDevice blackpixel == 0) == (photometric == #blackIs0)) ifFalse:[
                "have to invert bits"
                f function:#copyInverted
            ].
            aDevice drawBits:bytes depth:pictureDepth width:width height:height
                            x:0 y:0
                         into:(f id)
                            x:0 y:0 width:width height:height with:(f gcId).
            ^ f
        ].

        "the image has more greylevels than the display - dither"

"
coming soon ...
        DitherAlgorithm == #error ifTrue:[
            ^ self greyImageAsErrorDitheredGreyFormOn:aDevice
        ].
"
        DitherAlgorithm == #pattern ifTrue:[
            ^ self greyImageAsPatternDitheredGreyFormOn:aDevice
        ].

        "no dither, simply cut off information"
        (aDevice depth == 1) ifTrue:[
            "for monochrome, there is a special method to do this"
            ^ self greyImageAsMonoFormOn:aDevice
        ].
        "the general case, will take as many bits from the image
         as possible for the device"
        ^ self greyImageAsGreyFormOn:aDevice
    ].


    (aDevice visualType == #PseudoColor or:[aDevice visualType == #GrayScale]) ifTrue:[
        ^ self greyImageAsPseudoFormOn:aDevice
    ].

    self error:'cannot convert this format'.
    ^ nil
!

greyImageAsMonoFormOn:aDevice
    "return a (thresholded) monochrome Form from the picture."

    ^ self subclassResponsibility
!

greyImageAsPatternDitheredGreyFormOn:aDevice
    "return a dithered greyForm from the grey picture.
     Works for any source/destination depths, but very very slow
     since each pixel is processed individually.
     Usually redefined by subclasses for more performance"

    |f last      
     x0            "{Class: SmallInteger }"
     w             "{Class: SmallInteger }"
     h             "{Class: SmallInteger }"
     run           "{Class: SmallInteger }" |

    Transcript showCr:'slow dithering ..'. Transcript endEntry.

    w := width - 1.
    h := height - 1.

    "draw each pixel using dither color (let others do the dithering)
     although the code is simple, its very slow"

    f := Form width:width height:height depth:aDevice depth on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.

    0 to:h do:[:dstY |
        x0 := 0.
        run := 0.
        last := nil.
        self atY:dstY from:0 to:w do:[:x :srcColor |
            |dstColor|

            srcColor ~~ last ifTrue:[
                (run ~~ 0) ifTrue:[
                    f fillRectangleX:x0 y:dstY width:run height:1.
                ].
                run := 0.
                dstColor := Color grey:(srcColor greyIntensity).
                f paint:dstColor.
                last := srcColor.
                x0 := x
            ].
            run := run + 1
        ].
        f fillRectangleX:x0 y:dstY width:run height:1.
    ].
    ^ f
!

greyImageAsGreyFormOn:aDevice
    "return an 8-bit Form from the grey picture"

    |wideBits pictureDepth f map nplanes ncells
     inverse 
     mapSize  "{ Class: SmallInteger }"
     oldValue "{ Class: SmallInteger }"
     newValue "{ Class: SmallInteger }"
     shift    "{ Class: SmallInteger }"
     shift2   "{ Class: SmallInteger }"
     shift3   "{ Class: SmallInteger }"
     shift4   "{ Class: SmallInteger }" |

    (aDevice depth == 8) ifFalse:[
        'non-8 plane displays not supported' errorPrintNewline.
        ^ self greyImageAsMonoFormOn:aDevice
    ].

    pictureDepth := bitsPerSample at:1.

    wideBits := ByteArray uninitializedNew:(width * height).

    map := ByteArray uninitializedNew:(1 bitShift:pictureDepth).

    "find the real number of server-planes"
    nplanes := 8.
    ncells := 256.
    [aDevice ncells < ncells] whileTrue:[
        nplanes := nplanes - 1.
        ncells := ncells // 2
    ].

    "prepare translation table"

    shift := nplanes - pictureDepth.
    shift2 := shift - pictureDepth.
    shift3 := shift2 - pictureDepth.
    shift4 := shift3 - pictureDepth.

    inverse := aDevice blackpixel ~~ 0.
    photometric == #blackIs0 ifFalse:[
        inverse := inverse not
    ].
    mapSize := map size.
    1 to:mapSize do:[:index |
        oldValue := index - 1.
        newValue := oldValue bitShift:shift.
        newValue := newValue bitOr:(oldValue bitShift:shift2).
        newValue := newValue bitOr:(oldValue bitShift:shift3).
        newValue := newValue bitOr:(oldValue bitShift:shift4).
        inverse ifTrue:[
            map at:(map size - index + 1) put:newValue
        ] ifFalse:[
            map at:index put:newValue
        ]
    ].
    bytes expandPixels:pictureDepth
                width:width 
               height:height
                 into:wideBits
              mapping:map.

    f := Form width:width height:height depth:8 on:aDevice.
    f isNil ifTrue:[^ nil].
    f initGC.
    aDevice drawBits:wideBits depth:8 width:width height:height
                       x:0 y:0
                    into:(f id) x:0 y:0 width:width height:height with:(f gcId).
    ^ f
!

greyImageAsPseudoFormOn:aDevice
    "return an 8-bit pseudo Form from the grey picture"

    |wideBits pictureDepth f map  
     colorMap usedColors nUsed aColor 
     nColors "{ Class: SmallInteger }"
     range id|

    pictureDepth := bitsPerSample at:1.

    (#(2 4 8) includes:pictureDepth) ifFalse:[
        self error:'currently only depth-2, 4 or 8 supported'.
        ^ nil
    ].

    wideBits := ByteArray uninitializedNew:(width * height).

    (pictureDepth == 8) ifTrue:[
        "for 8bits, we scan for used colors first;
         to avoid allocating too many colors"

        usedColors := bytes usedValues.
        nUsed := usedColors maximumValue + 1.

        colorMap := Array new:nUsed.
        photometric == #blackIs0 ifTrue:[
            usedColors do:[:grey |
                colorMap at:(grey + 1) put:(Color grey:(100.0 / 256.0 * grey))
            ]
        ] ifFalse:[
            usedColors do:[:grey |
                colorMap at:(grey + 1) put:(Color grey:(100 - (100.0 / 256.0 * grey)))
            ]
        ]
    ] ifFalse:[
        nColors := (1 bitShift:pictureDepth).
        colorMap := Array new:nColors.
        range := 100 / (nColors - 1) asFloat.
        photometric == #blackIs0 ifTrue:[
            1 to:nColors do:[:i |
                colorMap at:i put:(Color grey:(i - 1) * range).
            ].
        ] ifFalse:[
            1 to:nColors do:[:i |
                colorMap at:(nColors - i + 1) put:(Color grey:(i - 1) * range).
            ].
        ].
    ].

    "XXX should reduce 8->6->4->2 planes, if not all colors could be allocated"

    "setup the translation map"
    map := ByteArray uninitializedNew:256.
    nColors := colorMap size.
    1 to:nColors do:[:i |
        aColor := colorMap at:i.
        aColor notNil ifTrue:[
            aColor := aColor on:aDevice.
            colorMap at:i put:aColor.
            id := aColor colorId.
            id notNil ifTrue:[
                map at:i put:id
            ] ifFalse:[
                map at:i put:0
            ]
        ]
    ].

    "expand & translate"
    bytes expandPixels:pictureDepth
                width:width 
               height:height
                 into:wideBits
              mapping:map.

    f := Form width:width height:height depth:8 on:aDevice.
    f isNil ifTrue:[^ nil].
    f colorMap:colorMap. 
    f initGC.
    aDevice drawBits:wideBits depth:8 width:width height:height
                       x:0 y:0
                    into:(f id) x:0 y:0 width:width height:height with:(f gcId).
    ^ f
! !

!Image methodsFor:'image manipulations'!

magnifyBy:extent
    "return a new image magnified by extent, aPoint.
     If non-integral magnify is asked for, pass the work on to 'hardMagnifyBy:'"

    |mX mY
     magX      "{ Class: SmallInteger }"   "new version of stc can find this out itself..."
     magY      "{ Class: SmallInteger }"
     srcOffset "{ Class: SmallInteger }"
     dstOffset "{ Class: SmallInteger }"
     w         "{ Class: SmallInteger }"
     h         "{ Class: SmallInteger }"
     first
     newWidth newHeight newImage newBits bitsPerPixel newBytesPerRow oldBytesPerRow|

    mX := extent x.
    mY := extent y.
    ((mX <= 0) or:[mY <= 0]) ifTrue:[^ nil].
    ((mX = 1) and:[mY = 1]) ifTrue:[^ self].

    ((mX isMemberOf:SmallInteger) and:[mY isMemberOf:SmallInteger]) ifFalse:[
        ^ self hardMagnifyBy:extent
    ].

    bitsPerPixel := self depth.
    oldBytesPerRow := ((width * bitsPerPixel) + 7) // 8.

    w := width.
    h := height.
    magX := mX.
    magY := mY.

    newWidth := w * mX.
    newHeight := h * mY.
    newBytesPerRow := ((newWidth * bitsPerPixel) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * newHeight).

    newImage := self species new.
    newImage bits:newBits.
    newImage width:newWidth.
    newImage height:newHeight.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.

    mX = 1 ifTrue:[
        "expand rows only"
        srcOffset := 1.
        dstOffset := 1.

        1 to:h do:[:row |
            1 to:mY do:[:i |
                newBits replaceFrom:dstOffset 
                        to:(dstOffset + oldBytesPerRow - 1)
                        with:bytes 
                        startingAt:srcOffset.
                dstOffset := dstOffset + newBytesPerRow
            ].
            srcOffset := srcOffset + oldBytesPerRow.
        ].
    ] ifFalse:[
        "expand cols"
        (mX > 1) ifTrue:[
            dstOffset := 1.
            srcOffset := 1.
            1 to:h do:[:row |
                self magnifyRowFrom:bytes 
                     offset:srcOffset  
                     into:newBits 
                     offset:dstOffset 
                     factor:mX.

                first := dstOffset.
                dstOffset := dstOffset + newBytesPerRow.
                " and copy for row expansion "
                2 to:mY do:[:i |
                    newBits replaceFrom:dstOffset 
                            to:(dstOffset + newBytesPerRow - 1)
                            with:newBits 
                            startingAt:first.
                    dstOffset := dstOffset + newBytesPerRow
                ].
                srcOffset := srcOffset + oldBytesPerRow.
            ].
        ]
    ].
    ^ newImage

    "((Image fromFile:'bitmaps/claus.gif') magnifyBy:1@2)"
!

hardMagnifyBy:extent
    "return a new image magnified by extent, aPoint.
     This is  the general magnification method, handling non-integral values"

    |mX        
     mY        
     newWidth  "{ Class: SmallInteger }"
     newHeight "{ Class: SmallInteger }"
     w         "{ Class: SmallInteger }"
     h         "{ Class: SmallInteger }"
     newImage newBits bitsPerPixel newBytesPerRow
     value srcRow|

    mX := extent x.
    mY := extent y.
    ((mX < 0) or:[mY < 0]) ifTrue:[^ nil].
    ((mX = 1) and:[mY = 1]) ifTrue:[^ self].

    newWidth := (width * mX) truncated.
    newHeight := (height * mY) truncated.

    bitsPerPixel := self depth.
    newBytesPerRow := ((newWidth * bitsPerPixel) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * newHeight).

    newImage := self species new.
    newImage bits:newBits.
    newImage width:newWidth.
    newImage height:newHeight.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.

    "walk over destination image fetching pixels from source image"

    w := newWidth - 1.
    h := newHeight - 1.

    0 to:h do:[:row |
        srcRow := (row // mY).
        0 to:w do:[:col |
            value := self valueAtX:(col // mX) y:srcRow.
            newImage atX:col y:row putValue:value.
        ]
    ].

    ^ newImage

    "((Image fromFile:'bitmaps/claus.gif') magnifyBy:0.5@0.5)"
!

flipHorizontal
    "inplace horizontal flip"

    |w  "{Class: SmallInteger }"
     h  "{Class: SmallInteger }"
     c2 "{Class: SmallInteger }" 
     value |

    w := width - 1.
    h := height - 1.

    0 to:h do:[:row |
        c2 := w.
        0 to:(w // 2) do:[:col |
            value := self valueAtX:col y:row.
            self atX:col y:row putValue:(self valueAtX:c2 y:row).
            self atX:c2 y:row putValue:value.
            c2 := c2 - 1.
        ]
    ].
    "flush device info"
    self restored
!

flipVertical
    "inplace vertical flip"

    |h           "{Class: SmallInteger }"
     bytesPerRow "{Class: SmallInteger }"
     buffer 
     indexLow    "{Class: SmallInteger }"
     indexHi     "{Class: SmallInteger }"|

    bytesPerRow := self bytesPerRow.
    buffer := ByteArray new:bytesPerRow.

    h := height - 1.

    indexLow := 1.
    indexHi := bytesPerRow * h + 1.

    0 to:(h // 2) do:[:row |
        buffer replaceFrom:1 to:bytesPerRow with:bytes startingAt:indexLow.
        bytes replaceFrom:indexLow to:(indexLow + bytesPerRow - 1) with:bytes startingAt:indexHi.
        bytes replaceFrom:indexHi to:(indexHi + bytesPerRow - 1) with:buffer startingAt:1.
        indexLow := indexLow + bytesPerRow.
        indexHi := indexHi - bytesPerRow.
    ].
    "flush device info"
    self restored
!

rotated:degrees
    "return a new image from the old one, by rotating the image
     degrees clockwise. 
     Currently, only rotation by a multiple of 90 degrees is implemented."

    |w  "{Class: SmallInteger }"
     h  "{Class: SmallInteger }"
     c2 "{Class: SmallInteger }" 
     newImage newBits newBytesPerRow d|

    d := degrees.
    [d < 0] whileTrue:[d := d + 360].
    d >= 360 ifTrue:[d := d \\ 360].
    d := d truncated.
    d == 0 ifTrue:[^ self].
    ((d ~~ 90) and:[(d ~~ 270) and:[d ~~ 180]]) ifTrue:[
        ^ self hardRotated:d
    ].

    newBytesPerRow := ((height * self depth) + 7) // 8.
    newBits := ByteArray uninitializedNew:(newBytesPerRow * width).

    newImage := self species new.
    newImage bits:newBits.
    newImage width:height.
    newImage height:width.
    newImage photometric:photometric.
    newImage samplesPerPixel:samplesPerPixel.
    newImage bitsPerSample:bitsPerSample.
    newImage colorMap:colorMap copy.

    w := width - 1.
    h := height - 1.

    d == 90 ifTrue:[
        0 to:h do:[:row |
            c2 := h-row.
            0 to:w do:[:col |
                newImage atX:c2 y:col putValue:(self valueAtX:col y:row).
            ]
        ]
    ].
    d == 180 ifTrue:[
        0 to:h do:[:row |
            0 to:w do:[:col |
                newImage atX:(h-row) y:(w-col) putValue:(self valueAtX:col y:row).
            ]
        ]
    ].
    d == 270 ifTrue:[
        0 to:h do:[:row |
            0 to:w do:[:col |
                newImage atX:row y:(w-col) putValue:(self valueAtX:col y:row).
            ]
        ]
    ].
    ^ newImage
!

hardRotated:degrees
    self error:'not yet implemented'
!

blurr
    "inplace blurr"

    "not yet implemented"

    "flush device info"
    self restored
! !

!Image methodsFor:'private'!

magnifyRowFrom:srcBytes offset:srcStart pixels:oldPixels 
          into:dstBytes offset:dstStart factor:mX

    "magnify a single pixel row - can only magnify by integer factors,
     can only magnify 1,2,4,8 and 24 bit-per-pixel images. But this is done fast."

    self subclassResponsibility
! !

!Image methodsFor: 'binary storage'!

readBinaryContentsFrom: stream manager: manager
    "tell the newly restored Image about restoration"

    super readBinaryContentsFrom: stream manager: manager.
    device := nil.
    deviceForm := nil.
    monoDeviceForm := nil.
    fullColorDeviceForm := nil.
! !
