'From Smalltalk/X, Version:1.5 on 6-Sep-91 at 10:37:07'!

Object subclass:#MailHandler
         instanceVariableNames:'mailFile userName headerStartPositions headerEndPositions
                headerLines mailFileSize tempDirName lockFileName canLock anyChanges
                mailDirName'
         classVariableNames:''
         poolDictionaries:''
         category:'Interface-MailTool'
!

MailHandler comment:'
 interface to mail-file

 %W% %E%
 written spring/summer 89 by claus

 I have put all interface to the mail files into this class,
 so that this code is not spread around in many mail-reader programs;
 currently, it can only extract header lines and individual mail
 entries - should be enhanced to allow for deleting / searching etc
'!

!MailHandler class methodsFor:'instance creation'!

new
    "answer a new mailhandler for users mail"

    ^ self basicNew initialize
! !

!MailHandler methodsFor:'initialization'!

initialize
    "initialize myself - scan through mailfile - get letter-positions"

    |pid|

    tempDirName := OperatingSystem getHomeDirectory , '/STmail'.
    (Disk isDirectory:tempDirName) ifFalse:[
        Disk createDirectory:tempDirName.
        (Disk isDirectory:tempDirName) ifFalse:[
            self notify:'cannot create temporary directory'.
            ^ self
        ]
    ].

    anyChanges := false.
    canLock := self setMailLock.
    self openMailFile.
    (mailFile notNil) ifTrue:[
        self scanMailFile
    ].
    canLock ifTrue:[
        self releaseMailLock
    ]
! !

!MailHandler methodsFor:'misc'!

openMailFile
    mailFile := self openMailFile:'' withMode:#readonly
!

createTemporaryNewMailFile
    ^ self openMailFile:'.tmp' withMode:#writing
!

openMailFile:extension withMode:how
    "try to open the mail file - it looks for it in '/usr/spool/mail and '/usr/mail'"

    | mailFileName stream |

    userName := OperatingSystem getLoginName.
    userName isNil ifTrue:[
        self error:'who are you'.
        ^ nil
    ].
    "try following mailboxes: $MAIL from environment,
                              /usr/spool/mail/<USERNAME>
                              /usr/mail/<USERNAME>
    "

    mailFileName := OperatingSystem getEnvironment:'MAIL'.
    mailFileName isNil ifTrue:[
        '/usr/spool/mail' asFilename exists ifTrue:[
            mailDirName := '/usr/spool/mail'
        ] ifFalse:[
            mailDirName := '/usr/mail'
        ].

        mailDirName asFilename exists ifFalse:[
            self warn:'no mail directory'.
            mailDirName := nil.
            ^ nil
        ].
        mailFileName := mailDirName , '/' , userName , extension.
    ].

    how == #readonly ifTrue:[
        stream := FileStream readonlyFileNamed:mailFileName.
    ] ifFalse:[
        stream := FileStream newFileNamed:mailFileName.
    ].
    ^ stream
!

release
    "release the mailbox - if any changes where made (i.e deletions),
     write back mailbox"

    mailFile notNil ifTrue:[
        mailFile close.
        mailFile := nil.
        OperatingSystem executeCommand:('rm -rf ' , tempDirName).
        (anyChanges and:[canLock]) ifTrue:[
            self updateMailFile
        ]
    ]
!

tempFilePath
    "answer the directory name where temporary files are stored
     i.e. where attachments where extracted"

    ^ tempDirName
!

scanMailFile
    "read mail file, fill startPosition and endPositions with From-line positions"

    |line|

    "no mail yet ? "
    mailFile isNil ifTrue:[self openMailFile].

    "still no mail"
    mailFile isNil ifTrue:[
        headerStartPositions := nil.
        headerEndPositions := nil.
        ^ self
    ].

    "size has not changed - just return"
    (mailFileSize == mailFile size) ifTrue:[^ self].

    "preallocating a bigger array avoids reallocating in grow ..."
    headerStartPositions := OrderedCollection new.

    mailFile position:1.
    mailFileSize := mailFile size.
    line := mailFile peekForLineStartingWith:'From '.
    [line notNil] whileTrue:[
        headerStartPositions add:(mailFile position).
        line := mailFile nextLine.
        line := mailFile peekForLineStartingWith:'From '
    ].

    headerEndPositions := OrderedCollection new.
    1 to:headerStartPositions size - 1 do:[:index |
        headerEndPositions add:(headerStartPositions at:(index + 1))
    ].
    headerEndPositions add:mailFileSize.
    headerLines := nil
!

updateMailFile
    "update the mailbox file"

    |again stream letter text mailFileName|

    [self setMailLock] whileFalse:[
        again := self confirm:'could not lock mail file - try again ?'.
        again ifFalse:[^ false].
    ].

    stream := self createTemporaryNewMailFile.
    stream isNil ifTrue:[
        self warn:'cannot create new mailbox file'.
        ^ false
    ].

    "simply append all left over mails ..."

    1 to: self numberOfLetters do:[:index |
        letter := self fullLetter:index.
        text := letter contents.
        text do:[:aLine |
            stream nextPutAll:aLine.
            stream cr.
        ]
    ].
    stream close.

    "now replace mailbox file by new file"

    mailFileName := mailDirName , '/' , OperatingSystem getLoginName.

    "during development of MailHandler - its a good idea to keep
     the previous mail in a backup - remove next line, when you think
     handler is ok"

    mailFileName asFilename renameTo:(mailFileName , '.bak').
    (mailFileName , '.tmp') asFilename renameTo:mailFileName.

    self releaseMailLock.
! !

!MailHandler methodsFor:'accessing'!

numberOfLetters
    "answer the number of letters in mail"

    self scanMailFile.
    headerStartPositions isNil ifTrue:[^ 0].
    ^ headerStartPositions size
!

headerLines
    "return an array filled with From-lines"

    |numberOfHeaderLines|

    headerLines notNil ifTrue:[ ^ headerLines ].

    headerStartPositions isNil ifTrue: [
        headerLines := Text with:'no mail'.
        ^ headerLines
    ].
    numberOfHeaderLines := headerStartPositions size.

    headerLines := Text new:numberOfHeaderLines.
    1 to:(numberOfHeaderLines) do:[:mailNr|
         headerLines at:mailNr put:(self summaryLineOf:mailNr)
    ].
    ^ headerLines
!

originOf:mailNr
    "extract origin of mail out of the From-line"

    |firstPosition lastPosition line
     fromLine address blankIndex moreToRead |

    headerStartPositions isNil ifTrue:[^ nil].
    firstPosition := headerStartPositions at:mailNr.
    lastPosition := headerEndPositions at:mailNr.

    mailFile position:firstPosition.
    line := mailFile nextLine.
    moreToRead := true.
    [moreToRead] whileTrue:[
        (line startsWith:'From ') ifTrue:[
            fromLine := line.
            moreToRead := false
        ] ifFalse:[
            line := mailFile nextLine.
            moreToRead := line notNil.
            ((mailFile position) >= lastPosition) ifTrue: [
                moreToRead := false
            ]
        ]
    ].
    fromLine isNil ifTrue:[^ nil].
    blankIndex := fromLine
                        indexOf:(Character space)
                        startingAt:6.
    (blankIndex == 0) ifTrue:[
        address := fromLine copyFrom:6 to:(fromLine size)
    ] ifFalse:[
        address := fromLine copyFrom:6 to:(blankIndex - 1)
    ].
    ^ address
!

summaryLineOf:mailNr
    "extract a header line consisting of: status, origin, date and subject"

    |fromLine fromLine2 subjectLine statusLine summaryLine
     address name dateString date status subject
     firstPosition lastPosition line
     index index2 moreToRead hasBeenRead|

    firstPosition := headerStartPositions at:mailNr.
    lastPosition := headerEndPositions at:mailNr.

    mailFile position:firstPosition.
    line := mailFile nextLine.
    moreToRead := true.
    [moreToRead] whileTrue:[
        (fromLine isNil and:[line startsWith:'From ']) ifTrue:[
            fromLine := line
        ] ifFalse:[
            (subjectLine isNil and:[line startsWith:'Subject: ']) ifTrue:[
                subjectLine := line
            ] ifFalse:[
                (statusLine isNil and:[line startsWith:'Status: ']) ifTrue:[
                    statusLine := line
                ] ifFalse: [
                    (fromLine2 isNil and:[line startsWith:'From: ']) ifTrue:[
                        fromLine2 := line
                    ]
                ]
            ]
        ].

        (fromLine notNil and:[subjectLine notNil]) ifTrue:[
            (statusLine notNil and:[fromLine2 notNil]) ifTrue:[
                "stop if we got all we need"
                moreToRead := false
            ]
        ].

        ((line size == 0) or:[line isBlank]) ifTrue:[
            "stop, when reaching end of header i.e. an empty line"

             moreToRead := false
        ].

        moreToRead ifTrue:[
            line := mailFile nextLine.
            moreToRead := line notNil.
            ((mailFile position) >= lastPosition) ifTrue: [
                moreToRead := false
            ]
        ]
    ].

    "now:
        fromLine:    first line matching 'From '
        fromLine2:                       'From: '
        subjectLine:                     'Subject: '
        statusLine:                      'Status: '
    "

    "extract date and email address from fromLine ...
     (format is 'From address date')"

    dateString := ''.
    fromLine isNil ifTrue:[
        address := 'BAD-ADDRESS'
    ] ifFalse:[
        index := fromLine indexOf:(Character space) startingAt:6.
        (index == 0) ifTrue:[
            address := fromLine copyFrom:6 to:(fromLine size)
        ] ifFalse:[
            address := fromLine copyFrom:6 to:(index - 1).
            dateString := fromLine copyFrom:(index + 1) to:(fromLine size)
        ]
    ].

    "extract real-life name from fromline2 (if present).
     some From:-lines contain full-name; format is varying (sigh)
     we recognize: 'From: '' full name '' ..... '
                   'From: full name <anything>'
                   'From: anything (full name)'
    "
    name := nil.
    fromLine2 notNil ifTrue:[
        "try name in double quotes"
        index := fromLine2 indexOf:(Character doubleQuote) startingAt:7.
        (index == 0) ifFalse:[
            index2 := fromLine2 indexOf:(Character doubleQuote)
                             startingAt:(index + 1).
            (index2 ~~ 0) ifTrue:[
                name := fromLine2 copyFrom:(index + 1) to:(index2 - 1)
            ]
        ].
        "no, look for < ... > take stuff before that"
        name isNil ifTrue:[
            index := fromLine2 indexOf:$< startingAt:7.
            (index == 0) ifFalse:[
                index2 := fromLine2 indexOf:$> startingAt:(index + 1).
                (index2 ~~ 0) ifTrue:[
                    name := fromLine2 copyFrom:7 to:(index - 1)
                ]
            ].

            name isNil ifTrue:[
                "no, try name in ( ... )"
                index := fromLine2 indexOf:$( startingAt:7.
                (index == 0) ifFalse:[
                    index2 := fromLine2 indexOf:$) startingAt:(index + 1).
                    (index2 ~~ 0) ifTrue:[
                        name := fromLine2 copyFrom:(index + 1) to:(index2 - 1)
                    ]
                ].

                name isNil ifTrue:[
                    "no, take rest of fromline2 as name"
                    name := fromLine2 copyFrom:7
                ]
            ]
        ]
    ].

    "no real name found - take reply address"
    name isNil ifTrue:[
        name := address
    ].

    name notNil ifTrue:[
        name := name withoutSpaces
    ].

    hasBeenRead := false.

    statusLine isNil ifTrue:[
        status := '  '
    ] ifFalse:[
        hasBeenRead := statusLine includes:$R
    ].

    subjectLine isNil ifTrue:[
        subject := ''
    ] ifFalse:[
        subject := subjectLine copyFrom:9 to:(subjectLine size)
    ].

    hasBeenRead ifFalse:[
        summaryLine := 'U '
    ] ifTrue:[
        summaryLine := '  '
    ].
    summaryLine := summaryLine , name , ' ' , dateString.
    subject notNil ifTrue:[
        summaryLine := summaryLine , ' '
                   , (Character doubleQuote asString)
                   , subject withoutSpaces
                   , (Character doubleQuote asString)
    ].
    ^ summaryLine
!

fullLetter:mailNr
    "read mail entry, return a MailLettercontaining all text (includes all relay stuff)"

    |firstPosition lastPosition line index lastSize lineArray 
     moreToRead hasAttachment ignore|

    headerStartPositions isNil ifTrue:[^ nil].

    firstPosition := headerStartPositions at:mailNr.
    lastPosition := headerEndPositions at:mailNr.

    "preallocating a bigger array avoids reallocating in grow ..."
    lineArray := Text new:100.
    lastSize := 100.
    lineArray grow:1.

    index := 1.
    mailFile position:firstPosition.
    line := mailFile nextLine.
    moreToRead := true.
    ignore := false.
    hasAttachment := false.

    [moreToRead] whileTrue:[
        ignore := false.

        (line startsWith:'Next-Attachment:') ifTrue:[
            hasAttachment := true
        ].

        hasAttachment ifTrue:[
            (line startsWith:'begin') ifTrue:[
                ((index + 3) > lastSize) ifTrue:[
                    lineArray grow:(lastSize * 2).
                    lastSize := lastSize * 2
                ].
                lineArray grow:(index + 3).
                lineArray at:index put:line.
                lineArray at:(index + 1) put:' ...'.
                lineArray at:(index + 2) put:'end'.
                index := index + 3.
                moreToRead := false.
                ignore := true
            ]
        ].

        ignore ifFalse:[
            lineArray at:index put:line.
            index := index + 1.
            (index > lastSize) ifTrue:[
                lineArray grow:(lastSize * 2).
                lastSize := lastSize * 2
            ].
            lineArray grow:index
        ].
        line := mailFile nextLine.
        line isNil ifTrue:[
            moreToRead := false
        ].
        ((mailFile position) > lastPosition) ifTrue: [
            moreToRead := false
        ]
    ].
    lineArray grow:(index - 1).
    ^ (MailLetter new type:#text) contents:lineArray
!

letter:mailNr
    "read mail entry, remove all unneeded text (inserted by relay-mailers),
     return a MailLetter containing the entry."

    |firstPosition lastPosition line lastSize index
     lineArray moreToRead inHeader ignoreThis attachmentLine 
     contentTypeLine contentEncodingLine boundaryString l
     tmpBaseName uuTmpFileName ZTmpFileName attachStream stream contents|

    headerStartPositions isNil ifTrue:[^ nil].

    firstPosition := headerStartPositions at:mailNr.
    lastPosition := headerEndPositions at:mailNr.

    "preallocating a bigger array avoids reallocating in grow ..."
    lineArray := Text new:100.
    lastSize := 100.
    lineArray grow:1.
    index := 1.

    attachmentLine := nil.
    contentTypeLine := nil.
    contentEncodingLine := nil.
    boundaryString := nil.
    mailFile position:firstPosition.                         
    inHeader := true.
    ignoreThis := true.
    line := mailFile nextLine.
    moreToRead := true.

    OperatingSystem executeCommand:('rm -rf ' , tempDirName).
    OperatingSystem executeCommand:('mkdir ' , tempDirName).

    tmpBaseName := tempDirName , '/letter'.
    uuTmpFileName := tmpBaseName , '.uu'.
    ZTmpFileName := tmpBaseName , '.Z'.

    [moreToRead] whileTrue:[
        ignoreThis := false.
        inHeader ifTrue:[
            ignoreThis := true.
            line isBlank ifTrue:[
                inHeader := false.
            ] ifFalse:[
                l := line asUppercase.
                (l startsWith:'NEXT-ATTACHMENT:') ifTrue:[
                    attachmentLine := line copyFrom:17
                ] ifFalse:[
                    (l startsWith:'CONTENT-TYPE:') ifTrue:[
                        contentTypeLine := line copyFrom:15.
                    ] ifFalse:[
                        (l startsWith:'CONTENT-TRANSFER-ENCODING:') ifTrue:[
                            contentEncodingLine := line copyFrom:28.
                        ]
                    ].
                ]
            ]
        ].

        ignoreThis ifFalse:[
            attachmentLine notNil ifTrue:[
                (line startsWith:'begin') ifTrue:[
                    attachStream := FileStream newFileNamed:uuTmpFileName.
                    attachStream notNil ifTrue:[
                        attachStream nextPutLine:('begin 666 ' , ZTmpFileName).
                        attachStream nextPutLinesFrom:mailFile upToLineStartingWith:'end'.
                        attachStream close.
                        ^ self unwrapNeXTAttachment:tmpBaseName inDirectory:tempDirName.
                    ]
                ]
            ].

            lineArray at:index put:line.
            index := index + 1.
            (index > lastSize) ifTrue:[
                lineArray grow:(lastSize * 2).
                lastSize := lastSize * 2
            ].
            lineArray grow:index
        ].

        line := mailFile nextLine.
        moreToRead := line notNil.

        ((mailFile position) > lastPosition) ifTrue: [
            moreToRead := false
        ]
    ].

    lineArray grow:(index - 1).
    contentTypeLine isNil ifTrue:[
        contentTypeLine := 'text/plain'
    ].

    (contentTypeLine startsWith:'text') ifTrue:[
        (contentTypeLine startsWith:'text/plain') ifTrue:[
            ^ self convertPlainDocument:lineArray encoding:contentEncodingLine
        ].
        (contentTypeLine startsWith:'text/richtext') ifTrue:[
            ^ self convertRichtextDocument:lineArray encoding:contentEncodingLine
        ].
        "default: present as-is"
        ^ (MailLetter new type:#text) contents:lineArray
    ].

    (contentTypeLine startsWith:'multipart') ifTrue:[
        ^ self unwrapMultipartDocument:lineArray
    ].

    (contentTypeLine startsWith:'image') ifTrue:[
        (contentTypeLine startsWith:'image/jpeg') ifTrue:[
            ^ self convertImageJPG:lineArray encoding:contentEncodingLine
        ].
        (contentTypeLine startsWith:'image/gif') ifTrue:[
            ^ self convertImageGIF:lineArray encoding:contentEncodingLine
        ].
        "default: present as-is"
        ^ (MailLetter new type:#text) contents:lineArray
    ].

    "default: present as-is"
    ^ (MailLetter new type:#text) contents:lineArray
!

deleteMail:mailNr
    "remove entry from list - should also remove in mailfile ..."

    |nLetters|

    headerStartPositions isNil ifTrue:[^ nil].
    nLetters := headerLines size.
    (mailNr > nLetters) ifTrue:[^ false].
    (nLetters > mailNr) ifTrue:[
        mailNr to:(nLetters - 1) do:[:index |
            headerStartPositions at:index put:(headerStartPositions at:(index + 1)).
            headerEndPositions at:index put:(headerEndPositions at:(index + 1)).
            headerLines at:index put:(headerLines at:(index + 1))
        ]
    ].
    headerStartPositions grow:(nLetters - 1).
    headerEndPositions grow:(nLetters - 1).
    headerLines grow:(nLetters - 1).
    anyChanges := true
!

sendMail:someText to:anAddress
    "send someText, aString to the destination anAddress, also a string"

    self sendMail:someText subject:nil cc:nil to:anAddress
!

subjectOf:mailNr
    "answer the subject of letter #mailNr"

    |firstPosition lastPosition line
     subjectLine address blankIndex moreToRead |

    headerStartPositions isNil ifTrue:[^ nil].
    firstPosition := headerStartPositions at:mailNr.
    lastPosition := headerEndPositions at:mailNr.

    mailFile position:firstPosition.
    line := mailFile nextLine.
    moreToRead := true.
    [moreToRead] whileTrue:[
        (line startsWith:'Subject: ') ifTrue:[
            subjectLine := line.
            moreToRead := false
        ] ifFalse:[
            line := mailFile nextLine.
            moreToRead := line notNil.
            ((mailFile position) >= lastPosition) ifTrue: [
                moreToRead := false
            ]
        ]
    ].
    subjectLine isNil ifTrue:[^ nil].
    ^ subjectLine copyFrom:10
!

sendMail:someText subject:subject cc:cc to:anAddress
    "send someText, aString to the destination anAddress, also a string"

    |stream|

    stream := PipeStream writingTo:('/usr/lib/sendmail ' , anAddress).
    stream isNil ifTrue:[
        self error:'cannot open pipe to mail'.
        ^ self
    ].

    (subject notNil and:[subject isBlank not]) ifTrue:[
        stream nextPutAll:('Subject: ' , subject).
        stream cr
    ].
    (cc notNil and:[cc isBlank not]) ifTrue:[
        stream nextPutAll:('CC: ' , cc).
        stream cr
    ].
    stream nextPutAll:someText.
    stream close
! !

!MailHandler methodsFor:'private'!

isMailerInfo:line
    "answer true, if the argument line is a mailer-info text line"

    |l|

    l := line asLowercase.
    (l startsWith:'in-reply-to: ') ifTrue:[^ true].
    (l startsWith:'apparently-to: ') ifTrue:[^ true].
    (l startsWith:'message-id: ') ifTrue:[^ true].
    (l startsWith:'mailed-from: ') ifTrue:[^ true].
    (l startsWith:'return-path: ') ifTrue:[^ true].
    (l startsWith:'posted-date: ') ifTrue:[^ true].
    (l startsWith:'cc: ') ifTrue:[^ true].
    (l startsWith:'x-mailer: ') ifTrue:[^ true].
    (l startsWith:'x-nx-transport-id: ') ifTrue:[^ true].
    (l startsWith:'sender: ') ifTrue:[^ true].
    (l startsWith:'status: ') ifTrue:[^ true].
    (l startsWith:'to: ') ifTrue:[^ true].
    (l startsWith:'from: ') ifTrue:[^ true].
    (l startsWith:'received: ') ifTrue:[^ true].
    (l startsWith:'via: ') ifTrue:[^ true].
    (l startsWith:'next-attachment: ') ifTrue:[^ true].
    (l startsWith:'mime-version: ') ifTrue:[^ true].
    (l startsWith:'content-type: ') ifTrue:[^ true].
    (l startsWith:'content-transfer-encoding: ') ifTrue:[^ true].
    (l startsWith:'content-length: ') ifTrue:[^ true].
    (l startsWith:'content-description: ') ifTrue:[^ true].
    (l startsWith:'keywords: ') ifTrue:[^ true].
    (l startsWith:'approved: ') ifTrue:[^ true].
    (l startsWith:'followup-to: ') ifTrue:[^ true].
    (l startsWith:'date: ') ifTrue:[^ true].
    (l = 'cc:') ifTrue:[^ true].
    ^ false
! !

!MailHandler methodsFor:'unwrapping'!

unwrapNeXTAttachment:baseName inDirectory:baseDir
    "unwrap all files in an attachment, return a letter containing the rtf-index file contents"

    |command ok stream contents|

    ok := true.
    Transcript showCr:'uudecoding ...'.
    command := 'cd ' , baseDir , '; uudecode < ' , baseName , '.uu'.
    (OperatingSystem executeCommand:command) ifFalse:[
        self warn:'cannot execute uudecode'.
        ok := false.
    ].

    ok ifTrue:[
        Transcript showCr:'uncompressing ...'.
        OperatingSystem executeCommand:('rm -f ' , baseName).
        command := 'cd ' , baseDir , '; uncompress ' , baseName , '.Z'.
        (OperatingSystem executeCommand:command) ifFalse:[
            self warn:'cannot execute uncompress'.
            ok := false
        ]
    ].

    ok ifTrue:[
        Transcript showCr:'untaring ...'.
        OperatingSystem isBSDlike ifTrue:[
            command := 'cd ' , baseDir , '; tar xf ' , baseName.
            (OperatingSystem executeCommand:command) ifFalse:[
                self warn:'tar cannot extract files'.
                ok := false
            ]
        ] ifFalse:[
            "on sys5 we use my tarx command (if available)
             - this one also extracts files with long names"
            command := 'cd ' , baseDir , '; tarx < ' , baseName.
            (OperatingSystem executeCommand:command) ifFalse:[
                command := 'cd ' , baseDir , '; tar xf ' , baseName.
                (OperatingSystem executeCommand:command) ifFalse:[
                    self warn:'tar cannot extract files'.
                    ok := false
                ]
            ]
        ].
    ].

    ok ifTrue:[
        stream := FileStream oldFileNamed:(baseDir , '/index.rtf').
        stream notNil ifTrue:[
            contents := stream contents.
            stream close.
            ^ (MailLetter new type:#rtf) contents:contents 
        ]
    ].

    ^ (MailLetter new type:#text) contents:#('failed to extract contents')
!

unwrapMultipartDocument:lines
    "no multipart unwrap at the moment - return the letters raw text for now"

    ^ (MailLetter new type:#text) contents:lines
! !

!MailHandler methodsFor:'converting'!

convertPlainDocument:lines encoding:encodingLine
    "plain text - just decode and return as text"

    ^ (MailLetter new type:#text) contents:(self decodeText:lines encoding:encodingLine)
!

convertRichtextDocument:lines encoding:encodingLine
    "richtext - just decode and return as text"

    ^ (MailLetter new type:#text) contents:(self decodeText:lines encoding:encodingLine)
!

convertImageJPG:lines encoding:encodingLine
    "no image support at the moment - return the letters raw text for now"

    ^ (MailLetter new type:#text) contents:lines
!

convertImageGIF:lines encoding:encodingLine
    "no image support at the moment - return the letters raw text for now"

    ^ (MailLetter new type:#text) contents:lines
! !

!MailHandler methodsFor:'decoding'!

decodeText:lines encoding:encodingLine
    encodingLine notNil ifTrue:[
        (encodingLine asLowercase startsWith:'quoted-printable') ifTrue:[
            ^ self decodeQuotedPrintable:lines
        ].
        (encodingLine asLowercase startsWith:'base64') ifTrue:[
            ^ (self decodeBase64:lines) asString
        ].
    ].
    "default: as is"
    ^ lines
!

decodeQuotedPrintable:lines
    |s in out skippingSpace spaceCount c hex1 hex2 hex val|

    s := lines asString.
    in := ReadStream on:s.
    out := WriteStream on:(String new:s size).
    skippingSpace := false.
    spaceCount := 0.
    [in atEnd] whileFalse:[
        c := in next.
        c == Character space ifTrue:[
            skippingSpace ifFalse:[
                spaceCount := spaceCount + 1.
            ]
        ] ifFalse:[
            "there is something"
            skippingSpace := false.
            c == Character cr ifTrue:[
                spaceCount := 0.
                out cr
            ] ifFalse:[
                spaceCount ~~ 0 ifTrue:[
                    out spaces:spaceCount.
                    spaceCount := 0
                ].
                (c == $=) ifTrue:[            "Rule 1"
                    hex1 := in next.
                    hex1 == Character cr ifTrue:[
                        "a soft line break"
                        spaceCount := 0
                    ] ifFalse:[
                        hex2 := in next.
                        hex2 == Character cr ifTrue:[
                            "half of a hex ?"
                            hex := hex1 asString.
                            val := Integer readFrom:(ReadStream on:hex) radix:16.
                            out nextPut:(Character value:val).
                            spaceCount := 0
                        ] ifFalse:[
                            hex := hex1 asString copyWith:hex2.
                            val := Integer readFrom:(ReadStream on:hex) radix:16.
                            out nextPut:(Character value:val)
                        ]
                    ]
                ] ifFalse:[
                    out nextPut:c
                ]
            ].
        ]
    ].
    ^ out contents asText
! 

decodeBase64:lines
    "see mime spec on whats going on here"

    |s in out 
     count "{ Class: SmallInteger }"
     bits  "{ Class: SmallInteger } - collects bits"
     b6    "{ Class: SmallInteger } - 6 encoded bits"
     pad atEnd c|

    s := lines asString.
    in := ReadStream on:s.
    out := WriteStream on:(ByteArray new:s size * 3 // 2).
    count := 0.
    bits := 0.
    atEnd := in atEnd.
    [atEnd] whileFalse:[
        c := in next.
        atEnd := in atEnd.
        b6 := nil.
        ((c >= $A) and:[c <= $Z]) ifTrue:[
            b6 := c asciiValue - $A asciiValue
        ] ifFalse:[
            ((c >= $a) and:[c <= $z]) ifTrue:[
                b6 := c asciiValue - $a asciiValue + 26
            ] ifFalse:[
                ((c >= $0) and:[c <= $9]) ifTrue:[
                    b6 := c asciiValue - $0 asciiValue + 52
                ] ifFalse:[
                    c == $+ ifTrue:[
                        b6 := 62
                    ] ifFalse:[
                        c == $/ ifTrue:[
                            b6 := 63
                        ]
                    ]
                ]
            ]
        ].
        b6 notNil ifTrue:[
            bits := (bits bitShift:6) + b6.
            count := count + 1.
            count == 4 ifTrue:[
                out nextPut:((bits bitShift:-16) bitAnd:16rFF).
                out nextPut:((bits bitShift:-8) bitAnd:16rFF).
                out nextPut:(bits bitAnd:16rFF).
                count := 0.
                bits := 0.

            ]
        ] ifFalse:[
            "other character - might be padding"
            c == $= ifTrue:[
                pad := 1.
                c := in next.
                c == $= ifTrue:[
                    pad := 2.
                ].
                atEnd := true
            ] ifFalse:[
                c == $- ifTrue:[
                    pad := 0.      "will add 0-bytes if not padded correctly"
                    atEnd := true
                ]
            ]
        ]
    ].
    count ~~ 0 ifTrue:[
        "at end, special considerations"

        [count ~~ 4] whileTrue:[
            bits := bits bitSHift:6.
            count := count + 1.
        ].
        pad == 2 ifTrue:[
            "only one valid byte"
            out nextPut:((bits bitShift:-16) bitAnd:16rFF).
        ] ifFalse:[
            pad == 1 ifTrue:[
                out nextPut:((bits bitShift:-16) bitAnd:16rFF).
                out nextPut:((bits bitShift:-8) bitAnd:16rFF).
            ] ifFalse:[
                "pad zero - add zero bytes"
                out nextPut:((bits bitShift:-16) bitAnd:16rFF).
                out nextPut:((bits bitShift:-8) bitAnd:16rFF).
                out nextPut:(bits bitAnd:16rFF).
            ]
        ]
    ].
    ^ out contents
! !

!MailHandler methodsFor:'locking'!

createLockLink:lockName from:tmpName
    "try to create a lock file, return true if successful.
     Only try for some time, then give up"

    10 timesRepeat:[
        (OperatingSystem linkFile:tmpName to:lockName) ifTrue:[^ true].
        Transcript showCr:'mail locked - please wait a bit'.
        Transcript endEntry.
        OperatingSystem sleep:1
    ].
    self warn:('could not create mail-lock; mailbox may only be read.\\or remove ' , lockName , ' if lock is a leftover.') withCRs.
    ^ false
!

setMailLock
    "create a lock file - stupid, mailers differ in how they lock,
     (I'm not even aware of how some of them do it ...).
     Return true if we have the lock, false if locking is not possible."

    |lockDir mailDir tmpName stream pid userName|

    userName := OperatingSystem getLoginName.

    (OperatingSystem getOSType = 'sunos') ifTrue:[
        "sunos uses a <username>.lock file in the mail-spool directory, which is empty"

        lockFileName := '/usr/spool/mail/' , userName , '.lock'.
        tmpName := '/usr/spool/mail/' , userName, '.' 
                                      , OperatingSystem getProcessId printString , '.tmp'.
        "create a temp file first ..."

        (FileStream newFileNamed:tmpName) isNil ifTrue:[
            self warn:'cannot create temporary for lock. Mailbox may only be read.'.
            ^ false
        ].

        (self createLockLink:lockFileName from:tmpName) ifFalse:[
            OperatingSystem removeFile:tmpName.
            self warn:'cannot greate lockfile. Mailbox may only be read.'.
            ^ false
        ].

        OperatingSystem removeFile:tmpName.
        ^ true
    ].

    "I cant remember if this works ..."
    (OperatingSystem getOSType = 'sco') ifTrue:[
        "sco uses a LCK..<pid> file in the mail-lock directory, which contains the pid"

        pid := OperatingSystem getProcessId printString.
        mailDir := FileDirectory directoryNamed:'/usr/mail'.
        mailDir isNil ifTrue:[
            self warn:'no mail directory'.
            ^ false
        ].

        lockDir := FileDirectory directoryNamed:'/usr/mail/:saved'.
        lockDir isNil ifTrue:[
            self warn:'no maillock directory. Mailbox may only be read.'.
            ^ false
        ].
        tmpName = 'LCK..' , pid.
        stream := FileStream newFileNamed:tmpName in:lockDir.
        stream isNil ifTrue:[
            self warn:'cannot create temporary lock file. Mailbox may only be read.'.
            ^ false
        ].
        stream nextPutAll:pid.
        stream close.

        lockFileName := userName , '.lock'.
        (self createLockLink:lockFileName from:('/usr/mail/:saved/' , tmpName)) ifFalse:[
            self warn:'cannot create lock file. Mailbox may only be read.'.
            ^ false
        ].
        lockDir removeFile:tmpName.
        ^ true
    ].

    ^ false
!

releaseMailLock
    OperatingSystem removeFile:lockFileName
! !
