"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#PersistencyManager
       instanceVariableNames:'dataBase nextSeq'
       classVariableNames:''
       poolDictionaries:''
       category:'System-Support'
!

PersistencyManager comment:'

COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

%W% %E%
'!

!PersistencyManager class methodsFor:'documentation'!

documentation
"
This class brings persistency to Smalltalk/X.
Instances of PersistencyManager allow storing & retrieval
of objects. A PersistencyManager is created with:

        pers := PersistencyManager file:'fileName'

it will create the persistency file if not already existing.
Objects are saved with:

        pers store:anObject
or:
        pers store:anObject key:aKey

in the first case, the persistencyManager will assign a new (unique)
key and return it; in the second case, the object is stored under aKey.
The key may be an Integer or a ByteArray-like object (i.e. String,
Symbol etc.). Storing nil is equivalent to deleting an object.

Objects are retrieved using:

        pers retrieve:key

which returns the object stored under key, or nil if nothing was stored
with this key.

Objects can be deleted with:

        pers delete:key

PersistencyManager uses DBFile and binaryStorage to store objects.
Since binaryStorage is not yet fully tested (& implemented), use this
goody with care. (for example, portability between platforms has not been
tested yet - especially, non-IEEE floats are not converted yet).

examples:

     'storing objects'
     |p|
     p := PersistencyManager file:'pers'.
     p store:#(1 2 3 4) key:1.
     p store:'hello world' key:2.
     p store:#(1 2 [1 2 3 4] [5 6 7 8] 9 10 $a 'hello') key:3.
     p store:1.23456 key:4.
     p store:#(1.23456 3.456789 98765.4) key:5.
     p store:(FloatArray withAll:#(1.23456 3.456789 98765.4)) key:6.
     p store:(DoubleArray withAll:#(1.23456 3.456789 98765.4)) key:7.

     'retrieving objects'
     |p|
     p := PersistencyManager file:'pers'.
     Transcript showCr:(p retrieve:1) displayString.
     Transcript showCr:(p retrieve:2) displayString.
     Transcript showCr:(p retrieve:3) displayString.
     Transcript showCr:(p retrieve:4) displayString.
     Transcript showCr:(p retrieve:5) displayString.
     Transcript showCr:(p retrieve:6) displayString.
     Transcript showCr:(p retrieve:7) displayString.

     'storing arrays'
     |p  a|
     p := PersistencyManager file:'pers'.
     a := (1 to:20) asArray.
     p store:a key:1.
     a := (1 to:20) asByteArray.
     p store:a key:2.
     a := (1 to:20) asFloatArray.
     p store:a key:3.
     a := (1 to:20) asDoubleArray.
     p store:a key:4.

     'retrieving'
     |p|
     p := PersistencyManager file:'pers'.
     (p retrieve:1) inspect.
     (p retrieve:2) inspect.
     (p retrieve:3) inspect.
     (p retrieve:4) inspect.

     'storing a more complex object (self referencing)'
     |p a d|
     p := PersistencyManager file:'pers'.
     a := Array new:3.
     a at:1 put:'hello'.
     a at:2 put:'self ref'.
     a at:3 put:a.
     d := Dictionary new.
     d at:#hello put:'hello'.
     d at:#world put:'world'.
     d at:'self-ref' put:a.
     p store:d key:1.

     'retrieving'
     |p|
     p := PersistencyManager file:'pers'.
     (p retrieve:1) inspect

     'storing a large object'
     |p coll b s|
     b := ByteArray new:100000.
     s := WriteStream on:b.

     p := PersistencyManager file:'pers'.
     coll := (1 to:10000) asArray.
     1 to:1000 do:[:i |
        coll at:i put:(i @ i)
     ].
     1000 to:2000 do:[:i |
        coll at:i put:(i -> i)
     ].
     2000 to:3000 do:[:i |
        coll at:i put:(Rectangle new)
     ].
     MessageTally spyOn:[  
         Time millisecondsToRun:[
             coll storeBinaryOn:s.
             p store:coll key:1.
         ]
     ]                     

     'retrieving'
     |p|
     p := PersistencyManager file:'pers'.
     (p retrieve:1) inspect

     |p|
     p := PersistencyManager file:'pers'.
     p store:(Image fromFile:'bitmaps/claus.gif') key:4.

     |p|
     p := PersistencyManager file:'pers'.
     (p retrieve:4) inspect.

     |p|
     p := PersistencyManager file:'pers'.
     Transcript showCr:(Time millisecondsToRun:[p store:'hello' key:100]).
     Transcript showCr:(Time millisecondsToRun:[p store:'world' key:101]).
     Transcript showCr:(Time millisecondsToRun:[p retrieve:100]         )

     'synchronous storing (i.e. every store goes into file);
      time on ELC is (via very busy NFS to SS10/30) about 60ms per stored object'
      time on HP is (via unloaded NFS to NeXT) about 11ms per stored object'
      time on NeXT is (to local disk) about 14ms per stored object'
     |p|
     Transcript showCr:(
        Time millisecondsToRun:[
            p := PersistencyManager file:'strings'.
            1 to:1000 do:[:key |
               |dataString|

               dataString := key printString.
               p store:dataString key:key.
            ].
        ]
     )

     'asynchronous storing (i.e. collects objects and writes from time to time);
      time on ELC is (via very busy NFS to SS10/30) about 22ms per stored object'
      time on HP is (via unloaded NFS to NeXT) about 5ms per stored object'
      time on NeXT is (to local disk) about 6ms per stored object'
     |p|
     Transcript showCr:(
        Time millisecondsToRun:[
            p := PersistencyManager file:'strings'.
            1 to:1000 do:[:key |
               |dataString|

               dataString := key printString.
               p store:dataString key:key sync:false.
            ].
            p sync
        ]
     )
     
     'retrieving;
      time on ELC is (via busy NFS to SS10/30) about 5ms per object'
      time on HP is (via unloaded NFS to NeXT) about 5ms per object'
      time on NeXT is (from local disk) about 5.6ms per object'
     |p|
     Transcript showCr:(
        Time millisecondsToRun:[
            p := PersistencyManager file:'strings'.
            1 to:1000 do:[:key |
               |dataString|

               dataString := p retrieve:key.
               dataString = key printString ifFalse:[
                   self error:'oops'
               ]
            ].
        ]
     )

     |p|
     p := PersistencyManager file:'strings'.
     (p retrieve:555) printNewline
"
! !

!PersistencyManager class methodsFor:'instance creation'!

file:aFileName
    "return a persistency manager storing objects in aFileName"

    ^ self new file:aFileName readOnly:false
! 

readOnlyFile:aFileName
    "return a persistency manager storing objects in aFileName"

    ^ self new file:aFileName readOnly:true 
! !

!PersistencyManager methodsFor:'private'!

file:aFileName readOnly:readonly 
    "open/create the persistency file;
     get/set the nextSeq key for automatic keying.
     (the nextSeq is stored under key 0 in the database)"

    |b|

    readonly ifTrue:[
        dataBase := DBFile readonlyDataBaseNamed:aFileName.
    ] ifFalse:[
        dataBase := DBFile dataBaseNamed:aFileName.
    ].
    dataBase isNil ifTrue:[
        self error:'cannot open database file: ' , aFileName.
        ^ nil
    ].
    (dataBase includesKey:#[0 0 0 0]) ifTrue:[
        b := dataBase at:#[0 0 0 0].
        nextSeq := b doubleWordAt:1.
    ] ifFalse:[
        nextSeq := 1.
    ]
!

makeKey:aKey
    "make a useful key from given key object. Also check for
     valid key (non-nil, integer ~~ 0, or ByteArray-like)"

    |keyBytes|

    (aKey isKindOf:ByteArray) ifTrue:[
        ^ aKey
    ].
    aKey isInteger ifTrue:[
        aKey ~= 0 ifTrue:[
            (aKey isMemberOf:SmallInteger) ifTrue:[
                keyBytes := ByteArray new:4.
                keyBytes doubleWordAt:1 put:aKey
            ] ifFalse:[
                keyBytes := aKey printString.
            ].
            ^ keyBytes
        ].
    ].
    self error:'invalid key'.
    ^ nil
! !

!PersistencyManager methodsFor:'queries'!

includesKey:aKey
    "return true, if data is stored under aKey, false otherwise"

    |keyBytes|

    keyBytes := self makeKey:aKey.
    keyBytes isNil ifTrue:[^ nil].

    ^ dataBase includesKey:keyBytes
! !

!PersistencyManager methodsFor:'misc'!

sync
    "flush buffered data to disk - needed after a
     store without sync"

    dataBase sync
! !

!PersistencyManager methodsFor:'deleting'!

delete:aKey
    "delete the object stored under aKey"

    self delete:aKey sync:true
!

delete:aKey sync:sync
    "delete the object stored under aKey. 
     if sync is true, data is flushed to disk; 
     otherwise stuff is cached in memory."

    |keyBytes|

    keyBytes := self makeKey:aKey.
    keyBytes isNil ifTrue:[^ nil].
    dataBase remove:keyBytes.
    sync ifTrue:[dataBase sync].
! !

!PersistencyManager methodsFor:'storing & retrieve'!

store:anObject
    "assign a key, store anObject; return the key"

    |key b|

    key := nextSeq.
    nextSeq := nextSeq + 1.
    "store the next-key"
    b := ByteArray new:4.
    b doubleWordAt:1 put:nextSeq.
    dataBase at:#[0 0 0 0] put:b.

    ^ self store:anObject key:key
!

store:anObject key:aKey
    "store anObject; return the key. The objects
     data is immediately synced to disk 
     (therefore this method may take some time)"

    ^ self store:anObject key:aKey sync:true
!

store:anObject key:aKey sync:sync
    "store anObject; return the key. if sync is true,
     data is flushed to disk; otherwise stuff is cached
     in memory."

    |s dataBytes keyBytes|

    keyBytes := self makeKey:aKey.
    keyBytes isNil ifTrue:[^ nil].

    anObject isNil ifTrue:[
        dataBase remove:keyBytes.
        dataBase sync.
    ] ifFalse:[
        s := ReadWriteStream on:(ByteArray new).
        anObject storeBinaryOn:s.
        dataBytes := s contents.

        dataBase at:keyBytes put:dataBytes.
        sync ifTrue:[dataBase sync].
    ].
    ^ aKey
!

retrieve:aKey
    "retrieve the object stored under aKey"

    |s dataBytes keyBytes|

    keyBytes := self makeKey:aKey.
    keyBytes isNil ifTrue:[^ nil].
    dataBytes := dataBase at:keyBytes.
    dataBytes isNil ifTrue:[^ nil].
    ^ (BinaryInputManager readFrom:(ReadStream on:dataBytes))
! !
