"       NAME            Paths
        AUTHOR          scaletti@uxc.cso.uiuc.edu (Kurt J. Hebel)
        FUNCTION        Reimplementation of the Path hierarchy
        ST-VERSION      2.2?
        PREREQUISITES   
        CONFLICTS       
        DISTRIBUTION    world
        VERSION         1
        DATE    3 Jun 1987
"

Path subclass: #Spline
        instanceVariableNames: 'function firstDerivative secondDerivative thirdDerivative'
        classVariableNames: ''
        poolDictionaries: ''
        category: 'Graphics-Paths'
!

Spline comment:
'Spline defines a path that includes an arbitrary collection of points connected
 by a third order curve.'
!

!Spline methodsFor: 'displaying'!

displayOn: aDisplayMedium at: aPoint clippingBox: clipRect rule: anInteger mask: aForm 
        "Display this Spline as a series of approximating lines.  Also, see comment
         in DisplayObject|displayOn:at:clippingBox:rule:mask:."

        | aLinearFit steps a b c d t |

        "Make sure that the function and its derivatives are up to date."
        self validateDerivatives.

        "Create a LinearFit for plotting."
        aLinearFit _ LinearFit new.
        aLinearFit form: self form.
        aLinearFit add: function first.

        "Approximate each spline knot."
        1 to: function size - 1 do: [ :k | 
                "Compute the Taylor series coefficients."
                d _ function at: k.
                c _ firstDerivative at: k.
                b _ (secondDerivative at: k) / 2.0.
                a _ (thirdDerivative at: k) / 6.0.

                "Compute the number of approximating segments."
                steps _ (secondDerivative at: k) abs + (secondDerivative at: k + 1) abs.
                steps _ 5 max: (steps x + steps y) // 100.

                "Add each of the approximating line segments."
                1 to: steps do: [ :j | 
                        t _ j asFloat / steps.
                        aLinearFit add: a * t + b * t + c * t + d].

                "Add the last line to the real spline endpoint."
                aLinearFit add: (function at: k + 1)].

        "Plot the LinearFit which approximates this Spline."
        aLinearFit
                displayOn: aDisplayMedium
                at: aPoint
                clippingBox: clipRect
                rule: anInteger
                mask: aForm.! !

!Spline methodsFor: 'private'!

computeCurve
        "Compute an array for the derivatives at each knot."
        | size values cyclic second third secondFromLast thirdFromLast |

        "Get the number of points, and make an OrderedColleciton of all of the points."

        size _ collectionOfPoints size.
        function _ OrderedCollection new: size.
        collectionOfPoints do: [ :point | function addLast: point].

        "Exit if the spline has only one point."
        size = 1 ifTrue: [^ self].

        "Flag whether curve is cyclic or not."
        cyclic _ size > 3 and: [function first = function last].

        "Set up the values collection.  The derivatives are computed from this."
        values _ function copy.

        "Process cyclic curves differently.  Add the last two points to the beginning,
         and the first two points to the end, so that the derivative calculation can
         look at a cycle."
        cyclic ifTrue: [
                second _ values at: 2.
                third _ values at: 3.
                thirdFromLast _ values at: size - 2.
                secondFromLast _ values at: size - 1.

                values addFirst: secondFromLast.
                values addFirst: thirdFromLast.
                values addLast: second.
                values addLast: third].

        "Compute the derivatives of the values collection."
        self computeDerivatives: values.

        "Remove any extra points which were added if the Spline is cyclic."
        cyclic ifTrue:  [
                firstDerivative removeFirst.
                firstDerivative removeFirst.
                firstDerivative removeLast.
                firstDerivative removeLast.
                secondDerivative removeFirst.
                secondDerivative removeFirst.
                secondDerivative removeLast.
                secondDerivative removeLast.
                thirdDerivative removeFirst.
                thirdDerivative removeFirst.
                thirdDerivative removeLast.
                thirdDerivative removeLast].!

computeDerivatives: values
        "Computes the first, second and third derivatives at each point in the collection
         values."
        | size v b lastV lastB nextV nextB valuesI valuesI1 valuesI2 twoDerivI twoDerivI1 |

        "Set up the derivative arrays."
        size _ values size.
        firstDerivative _ Array new: size.
        secondDerivative _ Array new: size.
        thirdDerivative _ Array new: size.

        "Compute the second derivative of the values."
        size > 2 ifTrue: [
                lastV _ 4.0.
                lastB _ 6.0 * (values first - ((values at: 2) * 2.0) + (values at: 3)).
                v _ Array new: size.
                b _ Array new: size.
                v at: 1 put: lastV.
                b at: 1 put: lastB.
                valuesI _ values at: 2.
                valuesI1 _ values at: 3.
                size > 3 ifTrue: [valuesI2 _ values at: 4].

                2 to: size - 2 do: [ :i |
                        nextV _ 4.0 - (1.0 / lastV).
                        nextB _ 6.0 * (valuesI - (valuesI1 * 2.0) + valuesI2) - (lastB / lastV).
                        v at: i put: nextV.
                        b at: i put: nextB.

                        size - 2 = i ifFalse: [
                                lastV _ nextV.
                                lastB _ nextB.
                                valuesI _ valuesI1.
                                valuesI1 _ valuesI2.
                                valuesI2 _ values at: i + 3]].

                secondDerivative at: size - 1 put: nextB / nextV.

                size - 2 to: 2 by: -1 do: [ :i | 
                        secondDerivative at: i put: (b at: i - 1)
                                                    - (secondDerivative at: i + 1) / (v at: i - 1)]].

        secondDerivative at: 1 put: 0.0 asPoint.
        secondDerivative at: size put: 0.0 asPoint.

        "Compute the values of the first and third derivative from the second derivative
         and the values."
        valuesI _ values at: 1.
        valuesI1 _ values at: 2.
        twoDerivI _ secondDerivative at: 1.
        twoDerivI1 _ secondDerivative at: 2.
        1 to: size - 1 do: [ :i |
                firstDerivative at: i put: valuesI1 - valuesI - (twoDerivI * 2.0 + twoDerivI1 / 6.0).
                thirdDerivative at: i put: twoDerivI1 - twoDerivI.

                size - 1 = i ifFalse: [
                        twoDerivI _ twoDerivI1.
                        twoDerivI1 _ secondDerivative at: i + 2.
                        valuesI _ valuesI1.
                        valuesI1 _ values at: i + 2]].

        "The derivative collections should be OrderedCollections."
        firstDerivative _ firstDerivative asOrderedCollection.
        secondDerivative _ secondDerivative asOrderedCollection.
        thirdDerivative _ thirdDerivative asOrderedCollection.!

validateDerivatives
        "Make sure that the function and derivative arrays are still valid.  If they
         are not, recompute them."
        | index |

        "Compute the derivatives if the cached function has not been computed."
        function isNil ifTrue: [^ self computeCurve].

        "Compute the derivatives if the cached function and the collection of points
         do not agree."
        index _ 1.
        collectionOfPoints do: [ :point |
                point ~~ (function at: index) ifTrue: [^ self computeCurve].
                index _ index + 1].! !

" The above file is a Manchester Goodie.  It is distributed freely on condition
 that you observe these conditions in respect of the whole Goodie, and on
 any significant part of it which is separately transmitted or stored:
        * You must ensure that every copy includes this notice, and that
          source and author(s) of the material are acknowledged.
        * These conditions must be imposed on anyone who receives a copy.
        * The material shall not be used for commercial gain without the prior
          written consent of the author(s).

 For more information about the Manchester Goodies Library (from which 
 this file was distributed) send e-mail:
        To: goodies-lib@cs.man.ac.uk
        Subject: help 
"!
